/*
Copyright (c) 2017, Intel Corp.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of Intel Corporation nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef IOCTL_H
#define IOCTL_H

#define _DRIVER_BUILD
#include "tdef.h"
#include <linux/ioctl.h>


#define SUCCESSFUL					0
#define INVALID_HANDLE				-1
#define MAJOR_NUM					253
#define DRIVER_NAME					"iommap"
#define DRIVER_DEVICE_FILE_NAME		"/dev/iommap"
#define DRIVER_PROC					"iMTA"
#define CMOS_SIZE					0xFF

// For 64bit address support
#ifdef x86_64
typedef unsigned long long	PCIADDR;
#else
typedef unsigned int		PCIADDR;
#endif

/********************************************************************************
 * Our user mode app will pass an initialized structure like this				*
 * down to the kernel mode driver												*
 ********************************************************************************/
typedef struct{
	UINT32 uBusNumber;
	UINT32 uSlotNumber;
	UINT32 uRegister;
	UINT32 uData;
} PCITEST_CONFIG_BLOCK,*PPCITEST_CONFIG_BLOCK;


/********************************************************************************
 * Common Input buffer for UCHAR												*
 * IOCTL_PCITEST_READ_PORT_UCHAR & IOCTL_PCITEST_WRITE_PORT_UCHAR				*
 * IOCTL_PCITEST_READ_REGISTER_UCHAR & IOCTL_PCITEST_WRITE_REGISTER_UCHAR		*
 ********************************************************************************/
typedef struct {
	UINT16 Port;
	UINT8 uData;
} PCITEST_PORT_UCHAR_BLOCK,*PPCITEST_PORT_UCHAR_BLOCK;

typedef struct {
	PCIADDR Register;
	UINT8 uData;
} PCITEST_REGISTER_UCHAR_BLOCK,*PPCITEST_REGISTER_UCHAR_BLOCK;


/********************************************************************************
 * Common Input buffer for USHORT												*
 * IOCTL_PCITEST_READ_PORT_USHORT & IOCTL_PCITEST_WRITE_PORT_USHORT				*
 * IOCTL_PCITEST_READ_REGISTER_USHORT & IOCTL_PCITEST_WRITE_REGISTER_USHORT		*
 ********************************************************************************/
typedef struct {
	UINT16 Port;
	UINT16 uData;
} PCITEST_PORT_USHORT_BLOCK,*PPCITEST_PORT_USHORT_BLOCK;

typedef struct {
	PCIADDR Register;
	UINT16 uData;
} PCITEST_REGISTER_USHORT_BLOCK,*PPCITEST_REGISTER_USHORT_BLOCK;


/********************************************************************************
 * Common Input buffer for ULONG												*
 * IOCTL_PCITEST_READ_PORT_ULONG & IOCTL_PCITEST_WRITE_PORT_ULONG				*
 * IOCTL_PCITEST_REGISTER_PORT_ULONG & IOCTL_PCITEST_WRITE_REGISTER_ULONG		*
 ********************************************************************************/
typedef struct {
	UINT16 Port;
	UINT32 uData;
} PCITEST_PORT_ULONG_BLOCK,*PPCITEST_PORT_ULONG_BLOCK;

typedef struct {
	PCIADDR Register;
	UINT32 uData;
} PCITEST_REGISTER_ULONG_BLOCK,*PPCITEST_REGISTER_ULONG_BLOCK;


/********************************************************************************
 * Common Input buffer for														*
 * IOCTL_PCITEST_READ_REGISTER_UCHAR_EX & IOCTL_PCITEST_WRITE_REGISTER_UCHAR_EX *
 * IOCTL_PCITEST_READ_PORT_USHORT_EX & IOCTL_PCITEST_WRITE_PORT_USHORT_EX		*
 * IOCTL_PCITEST_READ_PORT_ULONG_EX & IOCTL_PCITEST_WRITE_PORT_ULONG_EX			*
 ********************************************************************************/
typedef struct {
	UINT16 AddrPort;
	UINT16 DataPort;
	UINT8 uAddr;
	UINT8 uData;
} PCITEST_PORT_UCHAR_EX_BLOCK,*PPCITEST_PORT_UCHAR_EX_BLOCK;

typedef struct {
	UINT16 AddrPort;
	UINT16 DataPort;
	UINT16 uAddr;
	UINT16 uData;
} PCITEST_PORT_USHORT_EX_BLOCK,*PPCITEST_PORT_USHORT_EX_BLOCK;

typedef struct {
	UINT16 AddrPort;
	UINT16 DataPort;
	UINT32 uAddr;
	UINT32 uData;
} PCITEST_PORT_ULONG_EX_BLOCK,*PPCITEST_PORT_ULONG_EX_BLOCK;


/********************************************************************************
 * Common Input buffer for CMOSAPI_READ & CMOSAPI_WRITE							*
 ********************************************************************************/
typedef struct {
	UINT32  uStart;
	UINT32  uLength;
	UINT8  uData[CMOS_SIZE];
} CMOS_REQUEST,*PCMOS_REQUEST;


/********************************************************************************
 * Common Input buffer for MEMMAPAPI_GET_PHYSICAL_ADDRESS						*
 ********************************************************************************/
typedef struct {
	union {
		struct {
			UINT32 uLow;	//	low 32 Bit address 
			UINT32 uHigh;	//	high 32 Bit address
		};
		UINT64 uAddress;
	};
} PHYSICAL_ADDRESS, *PPHYSICAL_ADDRESS;


/********************************************************************************
 * Common Input buffer for VERSION_REQUEST										*
 ********************************************************************************/
typedef struct {
	UINT32  uVersion;
	CHAR  strVersion[0xFF];
	CHAR  strDescription[0xFF];
} VERSION_REQUEST,*PVERSION_REQUEST;


/********************************************************************************
 * Old version driver control code for backword compatibility support			*
 * For iommap driver version that less than 2.00.1.0803							*
 * Do not use this for new code!												*
 ********************************************************************************/
#define PCIAPI_READ_CONFIG_REG_OLD_REV _IOR(MAJOR_NUM, 0, PCITEST_CONFIG_BLOCK)
#define PCIAPI_WRITE_CONFIG_REG_OLD_REV _IOW(MAJOR_NUM, 1, PCITEST_CONFIG_BLOCK)

#define PCIAPI_READ_PORT_BYTE_OLD_REV _IOR(MAJOR_NUM, 2, PCITEST_PORT_UCHAR_BLOCK)
#define PCIAPI_WRITE_PORT_BYTE_OLD_REV _IOW(MAJOR_NUM, 3, PCITEST_PORT_UCHAR_BLOCK)
#define PCIAPI_READ_PORT_WORD_OLD_REV _IOR(MAJOR_NUM, 4, PCITEST_PORT_USHORT_BLOCK)
#define PCIAPI_WRITE_PORT_WORD_OLD_REV _IOW(MAJOR_NUM, 5, PCITEST_PORT_USHORT_BLOCK)
#define PCIAPI_READ_PORT_DWORD_OLD_REV _IOR(MAJOR_NUM, 6, PCITEST_PORT_ULONG_BLOCK)
#define PCIAPI_WRITE_PORT_DWORD_OLD_REV _IOW(MAJOR_NUM, 7, PCITEST_PORT_ULONG_BLOCK)

#define PCIAPI_READ_PORT_BYTE_EX_OLD_REV _IOR(MAJOR_NUM, 8, PCITEST_PORT_UCHAR_EX_BLOCK)
#define PCIAPI_WRITE_PORT_BYTE_EX_OLD_REV _IOW(MAJOR_NUM, 9, PCITEST_PORT_UCHAR_EX_BLOCK)
#define PCIAPI_READ_PORT_WORD_EX_OLD_REV _IOR(MAJOR_NUM, 10, PCITEST_PORT_USHORT_EX_BLOCK)
#define PCIAPI_WRITE_PORT_WORD_EX_OLD_REV _IOW(MAJOR_NUM, 11, PCITEST_PORT_USHORT_EX_BLOCK)
#define PCIAPI_READ_PORT_DWORD_EX_OLD_REV _IOR(MAJOR_NUM, 12, PCITEST_PORT_ULONG_EX_BLOCK)
#define PCIAPI_WRITE_PORT_DWORD_EX_OLD_REV _IOW(MAJOR_NUM, 13, PCITEST_PORT_ULONG_EX_BLOCK)

#define PCIAPI_READ_REGISTER_BYTE_OLD_REV _IOR(MAJOR_NUM, 14, PCITEST_REGISTER_UCHAR_BLOCK)
#define PCIAPI_WRITE_REGISTER_BYTE_OLD_REV _IOW(MAJOR_NUM, 15, PCITEST_REGISTER_UCHAR_BLOCK)
#define PCIAPI_READ_REGISTER_WORD_OLD_REV _IOR(MAJOR_NUM, 16, PCITEST_REGISTER_USHORT_BLOCK)
#define PCIAPI_WRITE_REGISTER_WORD_OLD_REV _IOW(MAJOR_NUM, 17, PCITEST_REGISTER_USHORT_BLOCK)
#define PCIAPI_READ_REGISTER_DWORD_OLD_REV _IOR(MAJOR_NUM, 18, PCITEST_REGISTER_ULONG_BLOCK)
#define PCIAPI_WRITE_REGISTER_DWORD_OLD_REV _IOW(MAJOR_NUM, 19, PCITEST_REGISTER_ULONG_BLOCK)

#define CMOSAPI_READ_OLD_REV _IOR(MAJOR_NUM, 20, CMOS_REQUEST)
#define CMOSAPI_WRITE_OLD_REV _IOW(MAJOR_NUM, 21, CMOS_REQUEST)

#define MEMMAPAPI_GET_PHYSICAL_ADDRESS_OLD_REV _IOR(MAJOR_NUM, 22, PHYSICAL_ADDRESS)

#define IOMMAP_GET_DRIVER_VERSION_OLD_REV _IOR(MAJOR_NUM, 23, VERSION_REQUEST)


/********************************************************************************
 * New version driver control code for backword compatibility support			*
 * For iommap driver version that above than 2.00.1.0803						*
 * Use this for new code!														*
 ********************************************************************************/
// For General Driver information API, reserve 0x00 to 0x1F
#define IOMMAP_GET_DRIVER_VERSION _IOR(MAJOR_NUM, 0x1F, VERSION_REQUEST)

// For PCI Config Register API, reserve 0x20 to 0x2F
#define PCIAPI_READ_CONFIG_REG_BYTE _IOR(MAJOR_NUM, 0x20, PCITEST_CONFIG_BLOCK)
#define PCIAPI_WRITE_CONFIG_REG_BYTE _IOW(MAJOR_NUM, 0x21, PCITEST_CONFIG_BLOCK)
#define PCIAPI_READ_CONFIG_REG_WORD _IOR(MAJOR_NUM, 0x22, PCITEST_CONFIG_BLOCK)
#define PCIAPI_WRITE_CONFIG_REG_WORD _IOW(MAJOR_NUM, 0x23, PCITEST_CONFIG_BLOCK)
#define PCIAPI_READ_CONFIG_REG_DWORD _IOR(MAJOR_NUM, 0x24, PCITEST_CONFIG_BLOCK)
#define PCIAPI_WRITE_CONFIG_REG_DWORD _IOW(MAJOR_NUM, 0x25, PCITEST_CONFIG_BLOCK)

// PCI Config Register API Backword Compatibility functions, do not use this for new code!
#define PCIAPI_READ_CONFIG_REG _IOR(MAJOR_NUM, 0x26, PCITEST_CONFIG_BLOCK)
#define PCIAPI_WRITE_CONFIG_REG _IOW(MAJOR_NUM, 0x27, PCITEST_CONFIG_BLOCK)

// For IO PORT API, reserve 0x30 to 0x3F
#define PCIAPI_READ_PORT_BYTE _IOR(MAJOR_NUM, 0x30, PCITEST_PORT_UCHAR_BLOCK)
#define PCIAPI_WRITE_PORT_BYTE _IOW(MAJOR_NUM, 0x31, PCITEST_PORT_UCHAR_BLOCK)
#define PCIAPI_READ_PORT_WORD _IOR(MAJOR_NUM, 0x32, PCITEST_PORT_USHORT_BLOCK)
#define PCIAPI_WRITE_PORT_WORD _IOW(MAJOR_NUM, 0x33, PCITEST_PORT_USHORT_BLOCK)
#define PCIAPI_READ_PORT_DWORD _IOR(MAJOR_NUM, 0x34, PCITEST_PORT_ULONG_BLOCK)
#define PCIAPI_WRITE_PORT_DWORD _IOW(MAJOR_NUM, 0x35, PCITEST_PORT_ULONG_BLOCK)

// For IO Index PORT API, reserve 0x40 to 0x4F
#define PCIAPI_READ_PORT_BYTE_EX _IOR(MAJOR_NUM, 0x40, PCITEST_PORT_UCHAR_EX_BLOCK)
#define PCIAPI_WRITE_PORT_BYTE_EX _IOW(MAJOR_NUM, 0x41, PCITEST_PORT_UCHAR_EX_BLOCK)
#define PCIAPI_READ_PORT_WORD_EX _IOR(MAJOR_NUM, 0x42, PCITEST_PORT_USHORT_EX_BLOCK)
#define PCIAPI_WRITE_PORT_WORD_EX _IOW(MAJOR_NUM, 0x43, PCITEST_PORT_USHORT_EX_BLOCK)
#define PCIAPI_READ_PORT_DWORD_EX _IOR(MAJOR_NUM, 0x44, PCITEST_PORT_ULONG_EX_BLOCK)
#define PCIAPI_WRITE_PORT_DWORD_EX _IOW(MAJOR_NUM, 0x45, PCITEST_PORT_ULONG_EX_BLOCK)

// For IO Memory API, reserve 0x50 to 0x5F
#define PCIAPI_READ_REGISTER_BYTE _IOR(MAJOR_NUM, 0x50, PCITEST_REGISTER_UCHAR_BLOCK)
#define PCIAPI_WRITE_REGISTER_BYTE _IOW(MAJOR_NUM, 0x51, PCITEST_REGISTER_UCHAR_BLOCK)
#define PCIAPI_READ_REGISTER_WORD _IOR(MAJOR_NUM, 0x52, PCITEST_REGISTER_USHORT_BLOCK)
#define PCIAPI_WRITE_REGISTER_WORD _IOW(MAJOR_NUM, 0x53, PCITEST_REGISTER_USHORT_BLOCK)
#define PCIAPI_READ_REGISTER_DWORD _IOR(MAJOR_NUM, 0x54, PCITEST_REGISTER_ULONG_BLOCK)
#define PCIAPI_WRITE_REGISTER_DWORD _IOW(MAJOR_NUM, 0x55, PCITEST_REGISTER_ULONG_BLOCK)

// For CMOS API, reserve 0x60 to 0x6F
#define CMOSAPI_KERNEL_READ _IOR(MAJOR_NUM, 0x60, CMOS_REQUEST)	// Do not use this for new code!
#define CMOSAPI_KERNEL_WRITE _IOW(MAJOR_NUM, 0x61, CMOS_REQUEST)// Do not use this for new code!
#define CMOSAPI_READ _IOR(MAJOR_NUM, 0x62, CMOS_REQUEST)
#define CMOSAPI_WRITE _IOW(MAJOR_NUM, 0x63, CMOS_REQUEST)

// For Memory API, reserve 0x70 to 0x7F
#define MEMMAPAPI_GET_PHYSICAL_ADDRESS _IOR(MAJOR_NUM, 0x70, PHYSICAL_ADDRESS)

#endif
