DEBUG = 0
TYPE = 32
OUTPUT = /usr/bin/
OBJDIR = obj$(TYPE)/
RELEASEDIR = Release$(TYPE)/
CC = gcc
LINK = ln

VERSION := $(shell grep " \bVERSION\b" main.h | sed -e 's/^\#.*	"//;s/".*//')

########################################################################################
#						for different /usr/src kernel source					       #
########################################################################################
KERNELDIR = $(shell find /usr/src/ -iname 'linux*' -type l -maxdepth 1)
INCLUDE = -I $(KERNELDIR)/include
WARN = -Wall

PARENTDIR = $(shell pwd)

########################################################################################
#							for different OS version								   #
########################################################################################
OSVER = $(shell cat /proc/version|sed -n '/SuSE/s/.*\(SuSE\).*/SUSE/p;')_def
ifneq "$(OSVER)" "SUSE_def"
	OSVER = UNDEF
else
	OSVER = SUSE
	
	# SuSE 8.2 use rpm -ba
	# SuSE 9.0 use rpmbuild -ba
	RPMCMD = $(shell find /usr/bin/ -iname 'rpmbuild')
	ifneq "$(RPMCMD)" ""
		RPMCMD = rpmbuild -ba
	else
		RPMCMD = rpm -ba
	endif

	RPMDIR = /usr/src/packages/
	SERVICECMD = /etc/init.d/
endif

ifeq "$(OSVER)" "UNDEF"
	OSVER = $(shell cat /proc/version|sed -n '/Red Hat/s/.*\(Red Hat\).*/REDHAT/p;')_def
	ifneq "$(OSVER)" "REDHAT_def"
		OSVER = UNDEF
	else
		OSVER = REDHAT
		RPMCMD = rpmbuild -ba
		RPMDIR = /usr/src/redhat/
		SERVICECMD = /sbin/service 
	endif
endif

ifeq "$(OSVER)" "UNDEF"
	OSVER = $(shell cat /proc/version|sed -n '/Gentoo/s/.*\(Gentoo\).*/GENTOO/p;')_def
	ifneq "$(OSVER)" "GENTOO_def"
		OSVER = UNDEF
		OSERR = $(error Unsupported OS !!! Currently support Redhat, SuSE and Gentoo)
	else
		OSVER = GENTOO
		RPMERR = $(error GENTOO is not supported RPM !!!)
		SERVICECMD = /etc/init.d/
	endif
endif


########################################################################################
#									for SMP											   #
########################################################################################
ifeq ($(KERNELDIR)/.config,$(wildcard $(KERNELDIR))/.config)
	include $(KERNELDIR)/.config
else
	MESSAGE := $(shell echo "WARNING: no .config file in $(KERNELDIR)")
endif


########################################################################################
#									for build										   #
########################################################################################
TARGET = iommap

SRC = main.c  pcitest.c cmos.c memmap.c
OBJ = $(OBJDIR)main.o $(OBJDIR)pcitest.o $(OBJDIR)cmos.o $(OBJDIR)memmap.o


MODCFLAGS += -D__KERNEL__ -DMODULE -D_$(OSVER)
MODCFLAGS += $(WARN) $(INCLUDE)

ifdef CONFIG_SMP
	MODCFLAGS += -D__SMP__ -DSMP
endif

#set MODCFALGS for optimization
#-02 works for Redhat8.2, Redhat9, AS3.0 and SuSE9
ifeq "$(DEBUG)" "0"
	ifeq "$(OSVER)" "REDHAT"
		MODCFLAGS += -O2
	else
		MODCFLAGS += -O
	endif
else
	MODCFLAGS += -g -O -D_DEBUG
endif


all:  $(OSERR) $(OBJDIR) $(OUTPUT) $(OUTPUT)$(TARGET).o 

rebuild: clean all


$(OBJDIR): 
	@if [ ! -d $(OBJDIR) ]; then mkdir $(OBJDIR); fi

$(OUTPUT):
	@if [ ! -d $(OUTPUT) ]; then mkdir $(OUTPUT); fi


$(OBJ): $(SRC)
	@echo "Compiling source files for OS:($(OSVER))..."
	@echo "Driver ver:($(TARGET)-$(VERSION))..."
	$(CC) $(MODCFLAGS) -c $(SRC)
	@mv -f $(SRC:.c=.o) $(OBJDIR).


$(OUTPUT)$(TARGET).o: $(OBJ)
	@echo "Linking binary files and copy installation files..."
	$(LD) -r $^ -o $@


clean:
	@echo "Cleaning object files, binaries and installation file..."
	@rm -rf *.o *.tar *.gz *.rpm
	@rm -rf $(OBJDIR)
	@rm -rf $(RELEASEDIR)
	@rm -rf $(OUTPUT)$(TARGET).o
	@rm -rf $(OUTPUT)$(TARGET).txt
	@rm -rf /dev/$(TARGET)
	@if [ -f /etc/init.d/iommapd ]; then $(SERVICECMD)iommapd unlink; fi
	@rm -rf /etc/init.d/iommapd
	@rm -rf /etc/iommapd.conf


conf: 
	@echo "Creating default Major/Minor number in /etc/iommapd.conf configuration files..."
	@echo "MAJOR=253" > /etc/iommapd.conf
	@echo "MINOR=0" >> /etc/iommapd.conf
	@echo "OSVER=$(OSVER)" >> /etc/iommapd.conf
	@echo "AUTO_DETECT=YES" >> /etc/iommapd.conf
	@chmod 644 /etc/iommapd.conf


install: clean all conf
	@echo "Installing driver:($(TARGET)) and service to run level 3 & 5..."
	@if [ ! -d /usr/local/man/man1 ]; then mkdir -p /usr/local/man/man1; fi
	@cp -f $(TARGET).txt $(OUTPUT)$(TARGET).txt
	@cd $(OUTPUT) && dos2unix $(OUTPUT)$(TARGET).txt && cd $(PARENTDIR)
	@cp -f iommapd /etc/init.d/iommapd
	@chmod +r /etc/init.d/iommapd
	@cd /etc/init.d && dos2unix /etc/init.d/iommapd && cd $(PARENTDIR)
	@chmod 755 /etc/init.d/iommapd
	@$(SERVICECMD)iommapd link
	@$(SERVICECMD)iommapd restart
	@echo "Updating Major number in iommapd.conf configuration files..."
	@grep "MAJOR=" /proc/iommap > /etc/iommapd.conf
	@grep "MINOR=" /proc/iommap >> /etc/iommapd.conf
	@echo "OSVER=$(OSVER)" >> /etc/iommapd.conf
	@echo "AUTO_DETECT=YES" >> /etc/iommapd.conf
	@chmod 644 /etc/iommapd.conf
	@$(SERVICECMD)iommapd reload


rpm: clean
	@echo "Creating RPM source and binary package file for $(TARGET) ver:($(VERSION))..."
	@if [ $(OSVER) = "GENTOO" ]; then echo $(RPMERR); fi
	@if [ -d /tmp/$(TARGET) ]; then rm -rf /tmp/$(TARGET); fi
	@cp -rHf $(PARENTDIR) /tmp
	@rm -rf /tmp/$(TARGET)/*.doc /tmp/$(TARGET)/*.xls
	@chmod +r /tmp/$(TARGET)/iommap.spec
	@cd /tmp && dos2unix /tmp/$(TARGET)/iommap.spec && cd $(PARENTDIR)
	@chmod 644 /tmp/$(TARGET)/iommap.spec
	@if [ -d /tmp/$(TARGET)-$(VERSION) ]; then rm -rf /tmp/$(TARGET)-$(VERSION); fi
	@mv -f /tmp/$(TARGET) /tmp/$(TARGET)-$(VERSION)
	@cd /tmp && tar -zvcf $(TARGET)-$(VERSION).tar.gz $(TARGET)-$(VERSION) && cd $(PARENTDIR)
	@mv -f /tmp/$(TARGET)-$(VERSION).tar.gz $(RPMDIR)SOURCES/
	@cd /tmp/$(TARGET)-$(VERSION) && $(RPMCMD) /tmp/$(TARGET)-$(VERSION)/$(TARGET).spec && cd $(PARENTDIR)
	@rm -rf /tmp/$(TARGET)-$(VERSION)


tar: clean
	@echo "Creating source and binary tar ball for $(TARGET) ver:($(VERSION))..."
	@if [ -d /tmp/$(TARGET) ]; then rm -rf /tmp/$(TARGET); fi
	@cp -rHf $(PARENTDIR) /tmp
	@rm -rf /tmp/$(TARGET)/*.doc /tmp/$(TARGET)/*.xls
	@chmod +r /tmp/$(TARGET)/iommap.spec
	@cd /tmp && dos2unix /tmp/$(TARGET)/iommap.spec && cd $(PARENTDIR)
	@chmod 644 /tmp/$(TARGET)/iommap.spec
	@if [ -d /tmp/$(TARGET)-$(VERSION) ]; then rm -rf /tmp/$(TARGET)-$(VERSION); fi
	@mv -f /tmp/$(TARGET) /tmp/$(TARGET)-$(VERSION)
	@cd /tmp && tar -zvcf $(TARGET)-$(VERSION)-src.tar.gz $(TARGET)-$(VERSION) && cd $(PARENTDIR)
	@if [ -d $(RELEASEDIR) ]; then rm -rf $(RELEASEDIR); fi
	@mkdir $(RELEASEDIR);
	@mv -f /tmp/$(TARGET)-$(VERSION)-src.tar.gz $(PARENTDIR)/$(RELEASEDIR)
	@rm -rf /tmp/$(TARGET)-$(VERSION)


dist: clean
	@echo "Creating release package file for $(TARGET) ver:($(VERSION))..."
	@if [ $(OSVER) = "REDHAT" -o $(OSVER) = "SUSE" ]; then { \
		make rpm; \
		if [ ! -d $(RELEASEDIR) ]; then { \
			mkdir $(RELEASEDIR); \
		} fi; \
		cp $(RPMDIR)SRPMS/$(TARGET)-$(VERSION)*.rpm $(RELEASEDIR); \
		find $(RPMDIR)RPMS -iname '$(TARGET)-$(VERSION)*.rpm' -exec cp {} $(RELEASEDIR) \; ; \
	} fi

	@if [ $(OSVER) = "GENTOO" ]; then { \
		make tar; \
		make all; \
		if [ ! -d $(RELEASEDIR) ]; then { \
			mkdir $(RELEASEDIR); \
		} fi; \
		cp -f $(OUTPUT)$(TARGET).o $(RELEASEDIR); \
		cp -f iommapd $(RELEASEDIR); \
		cp -f iommap_load $(RELEASEDIR); \
		cp -f $(TARGET).txt $(RELEASEDIR); \
		cd $(RELEASEDIR); \
		chmod +r iommapd; \
		chmod +r iommap_load; \
		dos2unix iommapd; \
		dos2unix iommap_load; \
		chmod 755 iommapd; \
		chmod 755 iommap_load; \
		tar zcvf $(TARGET)-$(VERSION)-bin.tar.gz * --exclude $(TARGET)-$(VERSION)-src.tar.gz; \
		./iommap_load install; \
		cd $(PARENTDIR); \
	} fi



.PHONY: clean all dist rebuild rpm
