/*
Copyright (c) 2017, Intel Corp.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of Intel Corporation nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>

#include <linux/version.h>
#include <linux/sched.h>
#if LINUX_VERSION_CODE > KERNEL_VERSION(4,11,99)
#include <linux/uaccess.h>		/* copy_to_user	*/
#else
#include <asm/uaccess.h>		/* copy_to_user	*/
#endif
#include <linux/fs.h>			/* file operation */
#include <linux/ioctl.h>		/* ioctl function */
#include <linux/proc_fs.h>		/* proc_dir_entry */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) & LINUX_VERSION_CODE < KERNEL_VERSION(2,7,0)
#include <linux/moduleparam.h>
#endif

#ifdef CONFIG_COMPAT
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
#include <asm/ioctl32.h>		/* ioctl_trans */
#endif
#endif

/******************************************************************************
 * header for driver                                                          *
 ******************************************************************************/
#include "wrapper.h"
#include "ioctl.h"

/******************************************************************************
 * Init Kernel driver                                                         *
 ******************************************************************************/
#define IOMMAP_VERSION     "2.09.1.0528"
#define IOMMAP_DESCRIPTION "Intel(R) PCI/IO/MEM Function Driver " \
			   "(version: "IOMMAP_VERSION", " \
			   "machine: "MACHINE", " \
			   "build: "BUILD")"
#define IOMMAP_COPYRIGHT   "Copyright(c) 2003-2017 Intel Corporation"
#define IOMMAP_LICENSE     "GPL"
#define IOMMAP_STRING_VER  "version: "IOMMAP_VERSION
#define IOMMAP_DWORD_VER   0x209

MODULE_DESCRIPTION(IOMMAP_DESCRIPTION);
MODULE_VERSION(IOMMAP_VERSION);
MODULE_AUTHOR(IOMMAP_COPYRIGHT);
MODULE_LICENSE(IOMMAP_LICENSE);

/******************************************************************************
 * External Interfaces                                                        *
 ******************************************************************************/
extern BOOL PciTestDispatch(unsigned int pciapiIoctl, unsigned long pciapiInfo);
extern BOOL CmosTestDispatch(unsigned int cmosapiIoctl, unsigned long cmosapiInfo);
extern BOOL MemMapTestDispatch(unsigned int memapiIoctl, unsigned long memmapapiInfo);

/******************************************************************************
 * Local Function Prototypes                                                  *
 ******************************************************************************/
static int driver_init(void);
static void driver_cleanup(void);
static int driver_open (struct inode *inode, struct file *file);
static int driver_release (struct inode *inode, struct file *file);
extern int driver_mmap(struct file *file, struct vm_area_struct *vma);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0)
static long driver_ioctl(struct file *file, unsigned int ioctl_num, unsigned long ioctl_param);
#else
static int driver_ioctl(struct inode *inode, struct file *file, unsigned int ioctl_num, unsigned long ioctl_param);
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,1,0)
ssize_t driver_proc_status(struct file *filp,char *buf,size_t count,loff_t *offp );
#else
static int driver_proc_status(char *page, char **start, off_t off, int count, int *eof, void *data);
#endif
module_init (driver_init);
module_exit (driver_cleanup);

#ifdef CONFIG_COMPAT
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
struct ioctl32_trans;
static int driver_ioctl32_map(unsigned int fd, unsigned int ioctl_num, unsigned long ioctl_param, struct file *file);
static int driver_ioctl32_register(struct ioctl32_trans *driver_map);
static int driver_ioctl32_unregister(struct ioctl32_trans *driver_map);
#else
static long driver_ioctl32_map(struct file *file, unsigned int ioctl_num, unsigned long ioctl_param);
#endif
#endif

/********************************************************************************
 * Structure declaration														*
 ********************************************************************************/
 #ifdef CONFIG_PROC_FS
static struct proc_dir_entry *driver_proc_entry = NULL;
struct proc_dir_entry *driver_proc_iMTA = NULL;
#endif /* CONFIG_PROC_FS */

static struct file_operations driver_fops = {
    .owner = THIS_MODULE,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0)
    .unlocked_ioctl = driver_ioctl,
#else
	.ioctl = driver_ioctl,
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,1,0)
	.read = driver_proc_status,
#endif
    .open = driver_open,

#ifdef _DEBUG
    .mmap = driver_mmap,
#endif

#ifdef CONFIG_COMPAT
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)
    .compat_ioctl = driver_ioctl32_map,
#endif
#endif

	.release = driver_release
};

#ifdef CONFIG_COMPAT
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
struct ioctl32_trans {
	unsigned long cmd;
	ioctl_trans_handler_t handler;
	int registered;
};
static struct ioctl32_trans ioctl32_map[]= {
	{PCIAPI_READ_CONFIG_REG, driver_ioctl32_map, 0},
	{PCIAPI_WRITE_CONFIG_REG, driver_ioctl32_map, 0},

	{PCIAPI_READ_CONFIG_REG_BYTE, driver_ioctl32_map, 0},
	{PCIAPI_WRITE_CONFIG_REG_BYTE, driver_ioctl32_map, 0},
	{PCIAPI_READ_CONFIG_REG_WORD, driver_ioctl32_map, 0},
	{PCIAPI_WRITE_CONFIG_REG_WORD, driver_ioctl32_map, 0},
	{PCIAPI_READ_CONFIG_REG_DWORD, driver_ioctl32_map, 0},
	{PCIAPI_WRITE_CONFIG_REG_DWORD, driver_ioctl32_map, 0},

	{PCIAPI_READ_PORT_BYTE, driver_ioctl32_map, 0},
	{PCIAPI_WRITE_PORT_BYTE, driver_ioctl32_map, 0},
	{PCIAPI_READ_PORT_WORD, driver_ioctl32_map, 0},
	{PCIAPI_WRITE_PORT_WORD, driver_ioctl32_map, 0},
	{PCIAPI_READ_PORT_DWORD, driver_ioctl32_map, 0},
	{PCIAPI_WRITE_PORT_DWORD, driver_ioctl32_map, 0},

	{PCIAPI_READ_PORT_BYTE_EX, driver_ioctl32_map, 0},
	{PCIAPI_WRITE_PORT_BYTE_EX, driver_ioctl32_map, 0},
	{PCIAPI_READ_PORT_WORD_EX, driver_ioctl32_map, 0},
	{PCIAPI_WRITE_PORT_WORD_EX, driver_ioctl32_map, 0},
	{PCIAPI_READ_PORT_DWORD_EX, driver_ioctl32_map, 0},
	{PCIAPI_WRITE_PORT_DWORD_EX, driver_ioctl32_map, 0},

	{PCIAPI_READ_REGISTER_BYTE, driver_ioctl32_map, 0},
	{PCIAPI_WRITE_REGISTER_BYTE, driver_ioctl32_map, 0},
	{PCIAPI_READ_REGISTER_WORD, driver_ioctl32_map, 0},
	{PCIAPI_WRITE_REGISTER_WORD, driver_ioctl32_map, 0},
	{PCIAPI_READ_REGISTER_DWORD, driver_ioctl32_map, 0},
	{PCIAPI_WRITE_REGISTER_DWORD, driver_ioctl32_map, 0},

	{CMOSAPI_READ, driver_ioctl32_map, 0},
	{CMOSAPI_WRITE, driver_ioctl32_map, 0},

	{MEMMAPAPI_GET_PHYSICAL_ADDRESS, driver_ioctl32_map, 0},

	{IOMMAP_GET_DRIVER_VERSION, driver_ioctl32_map, 0},
	{0, 0, 0},
};
#endif
#endif

/********************************************************************************
 * Driver version																*
 ********************************************************************************/
static char v[] = IOMMAP_STRING_VER;
static unsigned long g_Version	= IOMMAP_DWORD_VER;

/********************************************************************************
 * Global variable																*
 ********************************************************************************/
int AutoMajor_detect = 0;
static int g_Major = 0;
static int g_Minor = 0;
MODULE_PARM_INT(AutoMajor_detect, "i", 0644);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0)
DEFINE_SPINLOCK(g_iommapSpinLock);
#else
spinlock_t g_iommapSpinLock = SPIN_LOCK_UNLOCKED;
#endif

//static struct module this_module;

/********************************************************************************
 * Function	: driver_open														*
 * Access	: Local																*
 * Abstract	: Increase driver count if driver is opened. 						*
 ********************************************************************************/
static int driver_open(struct inode *inode,struct file *file)
{
	DMSG("FUNCTION: %s, LINE: %d, ARG(%p,%p)\n", __FUNCTION__, __LINE__, inode, file);

	IOMMAP_MODULE_INCREASE;

	return SUCCESSFUL;
}

/********************************************************************************
 * Function	: driver_release													*
 * Access	: Local																*
 * Abstract	: Decreade driver count if driver is released. 						*
 ********************************************************************************/
static int driver_release(struct inode *inode, struct file *file)
{
	DMSG("FUNCTION: %s, LINE: %d, ARG(%p,%p)\n", __FUNCTION__, __LINE__, inode, file);
 
	IOMMAP_MODULE_DECREASE;

	return SUCCESSFUL;
}

/********************************************************************************
 * Function	: driver_proc_status												*
 * Access	: Local																*
 * Abstract	: Output the driver information to /proc/iommap.					*
 ********************************************************************************/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,1,0)
ssize_t 
driver_proc_status(struct file *filp,char *buf,size_t count,loff_t *offp)
#else
static int 
driver_proc_status(char *page, char **start, off_t off, int count, int *eof, void *data)
#endif
{
	static char msg[256];
	static char tmp[256];
	int len = 0; 
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,1,0)
	static int finished = 0;
#endif
#ifdef _UBUNTU
	int i = 0;
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,1,0)
	DMSG("FUNCTION: %s, LINE: %d, ARG(%p,%p)\n", __FUNCTION__, __LINE__, buf, offp);
	if(finished)
	{
		finished = 0;
		return SUCCESSFUL;
	}
#else
	DMSG("FUNCTION: %s, LINE: %d, ARG(%p,%p)\n", __FUNCTION__, __LINE__, page, start);
	if(off > 0)
		return SUCCESSFUL;
#endif

	/* return driver version number, major and minor number to /proc/iommap */
	sprintf(tmp,"iommap driver %s (%s, %s)\nMAJOR=%d\nMINOR=%d", v, MACHINE, BUILD ,g_Major,g_Minor);
	len = sprintf(msg,"%s\n",tmp);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,1,0)
	finished = 1;
#ifdef _UBUNTU
	for(i=0; i < len; i++)
		put_user(msg[i], buf+i);
#else
	strcpy(buf, msg);
#endif
#else
	*eof = 1;
	strcpy(page, msg);
#endif
	printk("%s\n", msg);
	return len;
}

/********************************************************************************
 * Function	: driver_ioctl														*
 * Access	: Local																*
 * Abstract	: IO control code for the driver.									*
 ********************************************************************************/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0)
static long 
driver_ioctl(struct file *file, 
			 unsigned int ioctl_num, unsigned long ioctl_param)
#else
static int 
driver_ioctl(struct inode *inode, struct file *file, 
			 unsigned int ioctl_num, unsigned long ioctl_param)
#endif
{
	int status = EINVAL;
	VERSION_REQUEST driver_version;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0)
	DMSG("FUNCTION: %s, LINE: %d, ARG(%p):0x%X, 0x%lX\n", __FUNCTION__, __LINE__, file, ioctl_num, ioctl_param);
#else
	DMSG("FUNCTION: %s, LINE: %d, ARG(%p,%p):0x%X, 0x%lX\n", __FUNCTION__, __LINE__, inode, file, ioctl_num, ioctl_param);
#endif

	// Switch according to the ioctl called
	switch (ioctl_num) {

	// For Backword Compatibility functions, do not use this for new code!
	case PCIAPI_READ_CONFIG_REG:		
		//status = PciTestReadConfigRegDword(ioctl_param);
		status = PciTestDispatch(ioctl_num, ioctl_param);
		break;

	// For Backword Compatibility functions, do not use this for new code!
	case PCIAPI_WRITE_CONFIG_REG:		
		//status = PciTestWriteConfigRegDword(ioctl_param);
		status = PciTestDispatch(ioctl_num, ioctl_param);
		break;


	// For PCI Config Register API, reserve 0x20 to 0x2F
	case PCIAPI_READ_CONFIG_REG_BYTE:		
		//status = PciTestReadConfigRegByte(ioctl_param);
		status = PciTestDispatch(ioctl_num, ioctl_param);
		break;

	case PCIAPI_WRITE_CONFIG_REG_BYTE:		
		//status = PciTestWriteConfigRegByte(ioctl_param);
		status = PciTestDispatch(ioctl_num, ioctl_param);
		break;

	case PCIAPI_READ_CONFIG_REG_WORD:		
		//status = PciTestReadConfigRegWord(ioctl_param);
		status = PciTestDispatch(ioctl_num, ioctl_param);
		break;

	case PCIAPI_WRITE_CONFIG_REG_WORD:		
		//status = PciTestWriteConfigRegWord(ioctl_param);
		status = PciTestDispatch(ioctl_num, ioctl_param);
		break;

	case PCIAPI_READ_CONFIG_REG_DWORD:		
		//status = PciTestReadConfigRegDword(ioctl_param);
		status = PciTestDispatch(ioctl_num, ioctl_param);
		break;

	case PCIAPI_WRITE_CONFIG_REG_DWORD:		
		//status = PciTestWriteConfigRegDword(ioctl_param);
		status = PciTestDispatch(ioctl_num, ioctl_param);
		break;


	// For IO PORT API, reserve 0x30 to 0x3F
	case PCIAPI_READ_PORT_BYTE:
		//status = PciTestReadPortByte(ioctl_param);
		status = PciTestDispatch(ioctl_num, ioctl_param);
		break;

	case PCIAPI_WRITE_PORT_BYTE:
		//status = PciTestWritePortByte(ioctl_param);
		status = PciTestDispatch(ioctl_num, ioctl_param);
		break;

	case PCIAPI_READ_PORT_WORD:
		//status = PciTestReadPortWord(ioctl_param);
		status = PciTestDispatch(ioctl_num, ioctl_param);
		break;

	case PCIAPI_WRITE_PORT_WORD:
		//status = PciTestWritePortWord(ioctl_param);
		status = PciTestDispatch(ioctl_num, ioctl_param);
		break;

	case PCIAPI_READ_PORT_DWORD:	
		//status = PciTestReadPortDword(ioctl_param);
		status = PciTestDispatch(ioctl_num, ioctl_param);
		break;

	case PCIAPI_WRITE_PORT_DWORD:	
		//status = PciTestWritePortDword(ioctl_param);
		status = PciTestDispatch(ioctl_num, ioctl_param);
		break;


	// For IO Index PORT API, reserve 0x40 to 0x4F
	case PCIAPI_READ_PORT_BYTE_EX:
		//status = PciTestReadPortByteEx(ioctl_param);
		status = PciTestDispatch(ioctl_num, ioctl_param);
		break;

	case PCIAPI_WRITE_PORT_BYTE_EX:
		//status = PciTestWritePortByteEx(ioctl_param);
		status = PciTestDispatch(ioctl_num, ioctl_param);
		break;

	case PCIAPI_READ_PORT_WORD_EX:
		//status = PciTestReadPortWordEx(ioctl_param);
		status = PciTestDispatch(ioctl_num, ioctl_param);
		break;

	case PCIAPI_WRITE_PORT_WORD_EX:
		//status = PciTestWritePortWordEx(ioctl_param);	
		status = PciTestDispatch(ioctl_num, ioctl_param);
		break;

	case PCIAPI_READ_PORT_DWORD_EX:
		//status = PciTestReadPortDwordEx(ioctl_param);	
		status = PciTestDispatch(ioctl_num, ioctl_param);
		break;

	case PCIAPI_WRITE_PORT_DWORD_EX:
		//status = PciTestWritePortDwordEx(ioctl_param);
		status = PciTestDispatch(ioctl_num, ioctl_param);
		break;


	// For IO Memory API, reserve 0x50 to 0x5F
	case PCIAPI_READ_REGISTER_BYTE:
		//status = PciTestReadRegisterByte(ioctl_param);
		status = PciTestDispatch(ioctl_num, ioctl_param);
		break;

	case PCIAPI_WRITE_REGISTER_BYTE:
		//status = PciTestWriteRegisterByte(ioctl_param);
		status = PciTestDispatch(ioctl_num, ioctl_param);
		break;

	case PCIAPI_READ_REGISTER_WORD:
		//status = PciTestReadRegisterWord(ioctl_param);
		status = PciTestDispatch(ioctl_num, ioctl_param);
		break;

	case PCIAPI_WRITE_REGISTER_WORD:
		//status = PciTestWriteRegisterWord(ioctl_param);
		status = PciTestDispatch(ioctl_num, ioctl_param);
		break;

	case PCIAPI_READ_REGISTER_DWORD:
		//status = PciTestReadRegisterDword(ioctl_param);
		status = PciTestDispatch(ioctl_num, ioctl_param);
		break;

	case PCIAPI_WRITE_REGISTER_DWORD:
		//status = PciTestWriteRegisterDword(ioctl_param);
		status = PciTestDispatch(ioctl_num, ioctl_param);
		break;


	// For CMOS API, reserve 0x60 to 0x6F
	case CMOSAPI_KERNEL_READ:
		//status = CmosKernelRead(ioctl_param);
		status = CmosTestDispatch(ioctl_num, ioctl_param);
		break;

	case CMOSAPI_KERNEL_WRITE:	
		//status = CmosKernelWrite(ioctl_param);
		status = CmosTestDispatch(ioctl_num, ioctl_param);
		break;

	case CMOSAPI_READ:
		//status = CmosTestRead(ioctl_param);
		status = CmosTestDispatch(ioctl_num, ioctl_param);
		break;

	case CMOSAPI_WRITE:	
		//status = CmosTestWrite(ioctl_param);
		status = CmosTestDispatch(ioctl_num, ioctl_param);
		break;


	// For Memory API, reserve 0x70 to 0x7F
	case MEMMAPAPI_GET_PHYSICAL_ADDRESS:	
		//status = MemMapTestGetPhysicalAddress(ioctl_param);
		status = MemMapTestDispatch(ioctl_num, ioctl_param);
		break;


	// For General Driver information API, reserve 0x00 to 0x1F
	case IOMMAP_GET_DRIVER_VERSION:	
		memset(&driver_version, 0, sizeof(driver_version));
		driver_version.uVersion = g_Version;
		strcpy(driver_version.strVersion, v);
		strcpy(driver_version.strDescription, MACHINE);
		status = copy_to_user((void*)ioctl_param, &driver_version, sizeof(VERSION_REQUEST));
		if(status != SUCCESSFUL) {
			printk (KERN_ERR "FUNCTION: %s, LINE: %d, ERROR: COPY_TO_USER FOR IOMMAP_GET_DRIVER_VERSION(%d)\n",__FUNCTION__, __LINE__, status);
			return EFAULT;
		}
		break;


	default:
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: INVALID CASE, CMD(%d)\n",__FUNCTION__, __LINE__, ioctl_num);
        return EINVAL;
  }

  return status;
}

#ifdef CONFIG_COMPAT
/********************************************************************************
 * Function	: driver_ioctl32_conversion											*
 * Access	: Local																*
 * Abstract	: Handle for the 32bits to 64bits conversion in x86_64.				*
 ********************************************************************************/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
static int 
driver_ioctl32_map(unsigned int fd, unsigned int ioctl_num, unsigned long ioctl_param, struct file *file)
#else
static long driver_ioctl32_map(struct file *file, unsigned int ioctl_num, unsigned long ioctl_param)
#endif /* endif LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16) */
{
	int status = EINVAL;
	mm_segment_t old_fs;

	DMSG("FUNCTION: %s, LINE: %d, ARG(%p)\n", __FUNCTION__, __LINE__, file);
	printk (KERN_WARNING "%s is called.\n",__FUNCTION__);
	printk (KERN_WARNING "This function may cause kernel memory leak,\n");
	printk (KERN_WARNING "if the function is not called properly.\n");
	printk (KERN_WARNING "We suggest you check this call again.\n");
	printk (KERN_WARNING "Or recompile libiommap in pure 64bits.\n");

	/* Switch according to the ioctl called */
	switch (ioctl_num) {

	case MEMMAPAPI_GET_PHYSICAL_ADDRESS:	
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: INVALID CASE, CMD(%d), NOT FOR 32->64BITS\n",__FUNCTION__, __LINE__, ioctl_num);
        return EINVAL;

	default:
		old_fs = get_fs();
		set_fs(KERNEL_DS);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0)
		status = file->f_op->unlocked_ioctl(file, ioctl_num, ioctl_param);
#else
		status = file->f_op->ioctl(file->f_dentry->d_inode, file, ioctl_num, ioctl_param);
#endif
		set_fs(old_fs);
		break;
	}

  return status;
}

/********************************************************************************
 * Function	: driver_ioctl32_register											*
 * Access	: Local																*
 * Abstract	: Register all the handle for the 32bits to 64bits					*
 *			  conversion in x86_64.												*
 ********************************************************************************/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
static int driver_ioctl32_register(struct ioctl32_trans *driver_map)
{
	int err = 0, status = 0;
	struct ioctl32_trans *m;

	DMSG("FUNCTION: %s, LINE: %d, ARG(%p)\n", __FUNCTION__, __LINE__, driver_map);

	for(m = driver_map; m->cmd; m++) {
		err = register_ioctl32_conversion(m->cmd, m->handler);
		status |= err;

		if(err == 0) 
			m->registered++;
	}

	return status;
}
#endif /* endif LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16) */

/********************************************************************************
 * Function	: driver_ioctl32_unregister											*
 * Access	: Local																*
 * Abstract	: Unregister all the handle for the 32bits to 64bits				*
 *			  conversion in x86_64.												*
 ********************************************************************************/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
static int driver_ioctl32_unregister(struct ioctl32_trans *driver_map)
{
	int  status = 0;
	struct ioctl32_trans *m;

	DMSG("FUNCTION: %s, LINE: %d, ARG(%p)\n", __FUNCTION__, __LINE__, driver_map);

	for(m = driver_map; m->cmd; m++) {
		if(m->registered)	
			unregister_ioctl32_conversion(m->cmd);

		m->registered = 0;
	}

	return status;
}
#endif /* endif LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16) */
#endif /* endif CONFIG_COMPAT */

/********************************************************************************
 * Function	: driver_init														*
 * Access	: Local																*
 * Abstract	: Register driver module as character device.						*
 *			  /sbin/insmod iommap.o (normal driver loading fix to #253)			*
 *			  /sbin/insmod iommap.o AutoMajor_detect=1 (detect free resource)	*
 ********************************************************************************/
static int driver_init(void)
{
	int status = 0;

	printk("iommap driver loading with AutoMajor_detect: %d...\n",AutoMajor_detect);

	if(AutoMajor_detect == 0) {
		/* Major Number is fixed to 253. */
		status = register_chrdev(MAJOR_NUM, DRIVER_NAME, &driver_fops);
		g_Major = MAJOR_NUM;
	} 
	else {
		/* Dynamic Major will be use to determite free resource in a system */
		status = register_chrdev(0, DRIVER_NAME, &driver_fops);
		g_Major = status;
	}

	if (status < 0) {
		printk ("register %s module failed :%d\n", DRIVER_NAME, status);
		return status;
	}
	else {
		printk ("register %s module success.\n", DRIVER_NAME);
		printk ("The major device number is %d.\n", g_Major);
		printk ("We suggest you create a driver device file:\n");
		printk ("mknod %s c %d 0\n", DRIVER_DEVICE_FILE_NAME, g_Major);
		printk ("The device file name is important, because\n");
		printk ("the libiommap program assumes that's the\n");
		printk ("file you'll use.\n");
	}

#ifdef CONFIG_COMPAT
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
	status = driver_ioctl32_register(ioctl32_map);
	if(status) {
		printk ("%s driver_ioctl32_register failed to register:%d\n", DRIVER_NAME, status);
		driver_ioctl32_unregister(ioctl32_map);
		if(status) {
			printk ("%s driver_ioctl32_unregister failed to unregister:%d\n", DRIVER_NAME, status);
		}
		unregister_chrdev(g_Major, DRIVER_NAME);
		return status;
	}
#endif
#endif

	/* register the proc */
	driver_proc_iMTA = proc_mkdir(DRIVER_PROC, 0);

	if (driver_proc_iMTA) {
		/* linux 7 support */
		#if LINUX_VERSION_CODE < KERNEL_VERSION(3,1,0)
		driver_proc_entry = create_proc_entry(DRIVER_NAME, S_IFREG | S_IRUGO, driver_proc_iMTA);
		driver_proc_entry->read_proc = driver_proc_status;
		#else
		driver_proc_entry = proc_create(DRIVER_NAME, S_IFREG | S_IRUGO, driver_proc_iMTA, &driver_fops);
		#endif	
		
		printk ("created /proc/%s/%s directory.\n", DRIVER_PROC, DRIVER_NAME);
	} 
	else {
		printk ("Unable to create /proc/%s directory.\n", DRIVER_PROC);
	}

	return 0;
}

/********************************************************************************
 * Function	: driver_cleanup													*
 * Access	: Local																*
 * Abstract	: Unregister driver module as character device.						*
 ********************************************************************************/
static void driver_cleanup(void)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
	int status = unregister_chrdev(g_Major, DRIVER_NAME);

	if (status < 0)
		printk("Error in module_unregister_chrdev: %d\n", status);
#else
	unregister_chrdev(g_Major, DRIVER_NAME);
#endif

	if (driver_proc_entry != NULL)
		remove_proc_entry(DRIVER_NAME, driver_proc_iMTA);

	if (driver_proc_iMTA != NULL)
		remove_proc_entry(DRIVER_PROC, 0);


#ifdef CONFIG_COMPAT
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
	status = driver_ioctl32_unregister(ioctl32_map);
	if(status) {
		printk ("%s driver_ioctl32_unregister failed to unregister:%d\n", DRIVER_NAME, status);
	}
#endif
#endif

	printk ("removed /proc/%s/%s directory.\n", DRIVER_PROC, DRIVER_NAME);
	printk ("unregister %s module major# %d success.\n",DRIVER_NAME ,g_Major);
}  
