/*
Copyright (c) 2017, Intel Corp.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of Intel Corporation nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef __KERNEL__
#define __KERNEL__
#endif

#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#include <linux/module.h>

#define __NO_VERSION__
#include <linux/version.h>

#if LINUX_VERSION_CODE > KERNEL_VERSION(4,11,99)
#include <linux/uaccess.h>		/* copy_to_user	*/
#else
#include <asm/uaccess.h>		/* copy_to_user	*/
#endif

#include <asm/io.h>
//#include <asm/system.h>		// for global save_flag

#include <linux/spinlock.h>
#include <linux/ioport.h>
#include <linux/pci.h>

/********************************************************************************
 * header for driver															*
 ********************************************************************************/
#include "wrapper.h"
#include "ioctl.h"

/********************************************************************************
 * Local Function Prototypes													*
 ********************************************************************************/
// PCI API Interface
static BOOL PciTestReadConfigRegByte(unsigned long pciapiInfo);
static BOOL PciTestWriteConfigRegByte(unsigned long pciapiInfo);
static BOOL PciTestReadConfigRegWord(unsigned long pciapiInfo);
static BOOL PciTestWriteConfigRegWord(unsigned long pciapiInfo);
static BOOL PciTestReadConfigRegDword(unsigned long pciapiInfo);
static BOOL PciTestWriteConfigRegDword(unsigned long pciapiInfo);

// IO PORT API Interface
static BOOL PciTestReadPortByte(unsigned long pciapiInfo);
static BOOL PciTestWritePortByte(unsigned long pciapiInfo);
static BOOL PciTestReadPortWord(unsigned long pciapiInfo);
static BOOL PciTestWritePortWord(unsigned long pciapiInfo);
static BOOL PciTestReadPortDword(unsigned long pciapiInfo);
static BOOL PciTestWritePortDword(unsigned long pciapiInfo);

// IO PORT INDEX API Interface
static BOOL PciTestReadPortByteEx(unsigned long pciapiInfo);
static BOOL PciTestWritePortByteEx(unsigned long pciapiInfo);
static BOOL PciTestReadPortWordEx(unsigned long pciapiInfo);
static BOOL PciTestWritePortWordEx(unsigned long pciapiInfo);
static BOOL PciTestReadPortDwordEx(unsigned long pciapiInfo);
static BOOL PciTestWritePortDwordEx(unsigned long pciapiInfo);

// IO MEMORY API Interface
static BOOL PciTestReadRegisterByte(unsigned long pciapiInfo);
static BOOL PciTestWriteRegisterByte(unsigned long pciapiInfo);
static BOOL PciTestReadRegisterWord(unsigned long pciapiInfo);
static BOOL PciTestWriteRegisterWord(unsigned long pciapiInfo);
static BOOL PciTestReadRegisterDword(unsigned long pciapiInfo);
static BOOL PciTestWriteRegisterDword(unsigned long pciapiInfo);

/********************************************************************************
 * Global Function Prototypes													*
 ********************************************************************************/
extern spinlock_t g_iommapSpinLock;

/********************************************************************************
 * Function	: PciTestDispatch													*
 * Access	: Global															*
 * Abstract	: Dispatcher function for all PCI/IO/Memory access.					*
 ********************************************************************************/
BOOL PciTestDispatch(unsigned int pciapiIoctl, unsigned long pciapiInfo)
{
	int status = EINVAL;

	// Switch according to the ioctl called
	switch (pciapiIoctl) {

	// For Backword Compatibility functions, do not use this for new code!
	case PCIAPI_READ_CONFIG_REG:		
		status = PciTestReadConfigRegDword(pciapiInfo);	
		break;

	// For Backword Compatibility functions, do not use this for new code!
	case PCIAPI_WRITE_CONFIG_REG:		
		status = PciTestWriteConfigRegDword(pciapiInfo);		
		break;


	// For PCI Config Register API, reserve 0x20 to 0x2F
	case PCIAPI_READ_CONFIG_REG_BYTE:		
		status = PciTestReadConfigRegByte(pciapiInfo);	
		break;

	case PCIAPI_WRITE_CONFIG_REG_BYTE:		
		status = PciTestWriteConfigRegByte(pciapiInfo);		
		break;

	case PCIAPI_READ_CONFIG_REG_WORD:		
		status = PciTestReadConfigRegWord(pciapiInfo);	
		break;

	case PCIAPI_WRITE_CONFIG_REG_WORD:		
		status = PciTestWriteConfigRegWord(pciapiInfo);		
		break;

	case PCIAPI_READ_CONFIG_REG_DWORD:		
		status = PciTestReadConfigRegDword(pciapiInfo);	
		break;

	case PCIAPI_WRITE_CONFIG_REG_DWORD:		
		status = PciTestWriteConfigRegDword(pciapiInfo);		
		break;


	// For IO PORT API, reserve 0x30 to 0x3F
	case PCIAPI_READ_PORT_BYTE:			
		status = PciTestReadPortByte(pciapiInfo);		
		break;

	case PCIAPI_WRITE_PORT_BYTE:		
		status = PciTestWritePortByte(pciapiInfo);		
		break;

	case PCIAPI_READ_PORT_WORD:			
		status = PciTestReadPortWord(pciapiInfo);		
		break;

	case PCIAPI_WRITE_PORT_WORD:		
		status = PciTestWritePortWord(pciapiInfo);		
		break;

	case PCIAPI_READ_PORT_DWORD:		
		status = PciTestReadPortDword(pciapiInfo);		
		break;

	case PCIAPI_WRITE_PORT_DWORD:		
		status = PciTestWritePortDword(pciapiInfo);		
		break;


	// For IO Index PORT API, reserve 0x40 to 0x4F
	case PCIAPI_READ_PORT_BYTE_EX:		
		status = PciTestReadPortByteEx(pciapiInfo);		
		break;

	case PCIAPI_WRITE_PORT_BYTE_EX:		
		status = PciTestWritePortByteEx(pciapiInfo);	
		break;

	case PCIAPI_READ_PORT_WORD_EX:		
		status = PciTestReadPortWordEx(pciapiInfo);		
		break;

	case PCIAPI_WRITE_PORT_WORD_EX:		
		status = PciTestWritePortWordEx(pciapiInfo);	
		break;

	case PCIAPI_READ_PORT_DWORD_EX:		
		status = PciTestReadPortDwordEx(pciapiInfo);	
		break;

	case PCIAPI_WRITE_PORT_DWORD_EX:	
		status = PciTestWritePortDwordEx(pciapiInfo);	
		break;


	// For IO Memory API, reserve 0x50 to 0x5F
	case PCIAPI_READ_REGISTER_BYTE:		
		status = PciTestReadRegisterByte(pciapiInfo);	
		break;

	case PCIAPI_WRITE_REGISTER_BYTE:	
		status = PciTestWriteRegisterByte(pciapiInfo);	
		break;

	case PCIAPI_READ_REGISTER_WORD:		
		status = PciTestReadRegisterWord(pciapiInfo);	
		break;

	case PCIAPI_WRITE_REGISTER_WORD:	
		status = PciTestWriteRegisterWord(pciapiInfo);	
		break;

	case PCIAPI_READ_REGISTER_DWORD:	
		status = PciTestReadRegisterDword(pciapiInfo);	
		break;

	case PCIAPI_WRITE_REGISTER_DWORD:	
		status = PciTestWriteRegisterDword(pciapiInfo);	
		break;

	default:
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: INVALID CASE FOR PCI ACCESS, CMD(%d)\n",__FUNCTION__, __LINE__, pciapiIoctl);
        return EINVAL;
	}

	return status;
}

/********************************************************************************
 * Function	: PciTestReadConfigRegByte											*
 * Access	: Local																*
 * Abstract	: Read PCI configuration register in Byte.							*
 ********************************************************************************/
static BOOL PciTestReadConfigRegByte(unsigned long pciapiInfo)
{
	struct pci_dev *pdev = NULL; int status; unsigned char data = 0;
	PCITEST_CONFIG_BLOCK Block;

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, pciapiInfo);

	status = copy_from_user( &Block, (void*)pciapiInfo, sizeof(PCITEST_CONFIG_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	if (!PCIBIOS_PRESENT()) {
		printk (KERN_DEBUG "FUNCTION:%s, LINE: %d, ERROR: PCI_BIOS_PRESENT. NO PCI BIOS PRESENCT.\n",__FUNCTION__, __LINE__);
		return EPERM;
	}

	pdev = PCI_FIND_SLOT(Block.uBusNumber, Block.uSlotNumber);

	if(pdev == NULL) {
		printk (KERN_DEBUG "FUNCTION:%s, LINE: %d, ERROR: PCI_FIND_SLOT. NO PCI DEV FOUND.\n",__FUNCTION__, __LINE__);
		return ENODEV;
	}

	pci_read_config_byte(pdev, Block.uRegister, &data);
	
	Block.uData = data;

	status = copy_to_user((void*)pciapiInfo, &Block, sizeof(PCITEST_CONFIG_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_TO_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	DMSG (KERN_INFO "Block.uData 0x%X\n",Block.uData);

    return SUCCESSFUL;
}

/********************************************************************************
 * Function	: PciTestWriteConfigRegByte											*
 * Access	: Local																*
 * Abstract	: Write PCI configuration register in Byte.							*
 ********************************************************************************/
static BOOL PciTestWriteConfigRegByte(unsigned long pciapiInfo)
{
	struct pci_dev *pdev = NULL; int status; unsigned char data = 0;
	PCITEST_CONFIG_BLOCK Block;

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, pciapiInfo);

	status = copy_from_user( &Block, (void*)pciapiInfo, sizeof(PCITEST_CONFIG_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	if (!PCIBIOS_PRESENT())	{
		printk (KERN_DEBUG "FUNCTION:%s, LINE: %d, ERROR: PCI_BIOS_PRESENT. NO PCI BIOS PRESENCT.\n",__FUNCTION__, __LINE__);
		return EPERM;
	}

	pdev = PCI_FIND_SLOT(Block.uBusNumber, Block.uSlotNumber);

	if(pdev == NULL) {
		printk (KERN_DEBUG "FUNCTION:%s, LINE: %d, ERROR: PCI_FIND_SLOT. NO PCI DEV FOUND.\n",__FUNCTION__, __LINE__);
		return ENODEV;
	}

	data = Block.uData;

	pci_write_config_byte(pdev, Block.uRegister, data);

	DMSG (KERN_INFO "uBlock.uData 0x%X\n",Block.uData);

    return SUCCESSFUL;
}



/********************************************************************************
 * Function	: PciTestReadConfigRegWord											*
 * Access	: Local																*
 * Abstract	: Read PCI configuration register in Word.							*
 ********************************************************************************/
static BOOL PciTestReadConfigRegWord(unsigned long pciapiInfo)
{
	struct pci_dev *pdev = NULL; int status; unsigned short data = 0;
	PCITEST_CONFIG_BLOCK Block;

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, pciapiInfo);

	status = copy_from_user( &Block, (void*)pciapiInfo, sizeof(PCITEST_CONFIG_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	if (!PCIBIOS_PRESENT()) {
		printk (KERN_DEBUG "FUNCTION:%s, LINE: %d, ERROR: PCI_BIOS_PRESENT. NO PCI BIOS PRESENCT.\n",__FUNCTION__, __LINE__);
		return EPERM;
	}

	pdev = PCI_FIND_SLOT(Block.uBusNumber, Block.uSlotNumber);

	if(pdev == NULL) {
		printk (KERN_DEBUG "FUNCTION:%s, LINE: %d, ERROR: PCI_FIND_SLOT. NO PCI DEV FOUND.\n",__FUNCTION__, __LINE__);
		return ENODEV;
	}

	pci_read_config_word(pdev, Block.uRegister, &data);

	Block.uData = data;

	status = copy_to_user((void*)pciapiInfo, &Block, sizeof(PCITEST_CONFIG_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_TO_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	DMSG (KERN_INFO "Block.uData 0x%X\n",Block.uData);

    return SUCCESSFUL;
}

/********************************************************************************
 * Function	: PciTestWriteConfigRegWord											*
 * Access	: Local																*
 * Abstract	: Write PCI configuration register in Word.							*
 ********************************************************************************/
static BOOL PciTestWriteConfigRegWord(unsigned long pciapiInfo)
{
	struct pci_dev *pdev = NULL; int status; unsigned short data = 0;
	PCITEST_CONFIG_BLOCK Block;

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, pciapiInfo);

	status = copy_from_user( &Block, (void*)pciapiInfo, sizeof(PCITEST_CONFIG_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	if (!PCIBIOS_PRESENT())	{
		printk (KERN_DEBUG "FUNCTION:%s, LINE: %d, ERROR: PCI_BIOS_PRESENT. NO PCI BIOS PRESENCT.\n",__FUNCTION__, __LINE__);
		return EPERM;
	}

	pdev = PCI_FIND_SLOT(Block.uBusNumber, Block.uSlotNumber);

	if(pdev == NULL) {
		printk (KERN_DEBUG "FUNCTION:%s, LINE: %d, ERROR: PCI_FIND_SLOT. NO PCI DEV FOUND.\n",__FUNCTION__, __LINE__);
		return ENODEV;
	}

	data = Block.uData;

	pci_write_config_word(pdev, Block.uRegister, data);

	DMSG (KERN_INFO "uBlock.uData 0x%X\n",Block.uData);

    return SUCCESSFUL;
}



/********************************************************************************
 * Function	: PciTestReadConfigRegDword											*
 * Access	: Local																*
 * Abstract	: Read PCI configuration register in Dword.							*
 ********************************************************************************/
static BOOL PciTestReadConfigRegDword(unsigned long pciapiInfo)
{
	struct pci_dev *pdev = NULL; int status;
	PCITEST_CONFIG_BLOCK Block;

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, pciapiInfo);

	status = copy_from_user( &Block, (void*)pciapiInfo, sizeof(PCITEST_CONFIG_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	if (!PCIBIOS_PRESENT()) {
		printk (KERN_DEBUG "FUNCTION:%s, LINE: %d, ERROR: PCI_BIOS_PRESENT. NO PCI BIOS PRESENCT.\n",__FUNCTION__, __LINE__);
		return EPERM;
	}

	pdev = PCI_FIND_SLOT(Block.uBusNumber, Block.uSlotNumber);

	if(pdev == NULL) {
		//printk (KERN_DEBUG "FUNCTION:%s, LINE: %d, ERROR: PCI_FIND_SLOT. NO PCI DEV FOUND.\n",__FUNCTION__, __LINE__);
		DMSG (KERN_DEBUG "FUNCTION:%s, LINE: %d, ERROR: PCI_FIND_SLOT. NO PCI DEV FOUND.\n",__FUNCTION__, __LINE__);
		return ENODEV;
	}

	pci_read_config_dword(pdev, Block.uRegister, &Block.uData);


	status = copy_to_user((void*)pciapiInfo, &Block, sizeof(PCITEST_CONFIG_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_TO_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	DMSG (KERN_INFO "Block.uData 0x%X\n",Block.uData);

    return SUCCESSFUL;
}

/********************************************************************************
 * Function	: PciTestWriteConfigRegDword										*
 * Access	: Local																*
 * Abstract	: Write PCI configuration register in Dword.						*
 ********************************************************************************/
static BOOL PciTestWriteConfigRegDword(unsigned long pciapiInfo)
{
	struct pci_dev *pdev = NULL; int status;
	PCITEST_CONFIG_BLOCK Block;

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, pciapiInfo);

	status = copy_from_user( &Block, (void*)pciapiInfo, sizeof(PCITEST_CONFIG_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	if (!PCIBIOS_PRESENT())	{
		printk (KERN_DEBUG "FUNCTION:%s, LINE: %d, ERROR: PCI_BIOS_PRESENT. NO PCI BIOS PRESENCT.\n",__FUNCTION__, __LINE__);
		return EPERM;
	}

	pdev = PCI_FIND_SLOT(Block.uBusNumber, Block.uSlotNumber);

	if(pdev == NULL) {
		printk (KERN_DEBUG "FUNCTION:%s, LINE: %d, ERROR: PCI_FIND_SLOT. NO PCI DEV FOUND.\n",__FUNCTION__, __LINE__);
		return ENODEV;
	}

	pci_write_config_dword(pdev, Block.uRegister, Block.uData);

	DMSG (KERN_INFO "uBlock.uData 0x%X\n",Block.uData);

    return SUCCESSFUL;
}



/********************************************************************************
 * Function	: PciTestReadPortByte												*
 * Access	: Local																*
 * Abstract	: Read IO port register in byte.									*
 ********************************************************************************/
static BOOL PciTestReadPortByte(unsigned long pciapiInfo)
{
	int status; unsigned long flags;
	PCITEST_PORT_UCHAR_BLOCK Block;

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, pciapiInfo);

	status = copy_from_user( &Block, (void*)pciapiInfo, sizeof(PCITEST_PORT_UCHAR_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	// disable global irq and spinlock
	//do { save_flags(flags);	cli(); spin_lock(&g_iommapSpinLock); }while (0);
	//spin_lock_irqsave(&g_iommapSpinLock, flags);
	XSPIN_LOCK(g_iommapSpinLock, flags);

	Block.uData = inb(Block.Port);

	// restore global irq and unlock
	//do { spin_unlock(&g_iommapSpinLock); restore_flags(flags); /*sti();*/}while (0);
	//spin_unlock_irqrestore(&g_iommapSpinLock, flags);
	XSPIN_UNLOCK(g_iommapSpinLock, flags);

	status = copy_to_user((void*)pciapiInfo, &Block, sizeof(PCITEST_PORT_UCHAR_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_TO_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	DMSG (KERN_INFO "Block.Port 0x%X, Block.uData 0x%X\n",(unsigned)Block.Port,Block.uData);
    
	return SUCCESSFUL;
}

/********************************************************************************
 * Function	: PciTestWritePortByte												*
 * Access	: Local																*
 * Abstract	: Write IO port register in byte.									*
 ********************************************************************************/
static BOOL PciTestWritePortByte(unsigned long pciapiInfo)
{
	int status; unsigned long flags;
	PCITEST_PORT_UCHAR_BLOCK Block;

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, pciapiInfo);

	status = copy_from_user( &Block, (void*)pciapiInfo, sizeof(PCITEST_PORT_UCHAR_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	// disable global irq and spinlock
	//do { save_flags(flags);	cli(); spin_lock(&g_iommapSpinLock); }while (0);
	//spin_lock_irqsave(&g_iommapSpinLock, flags);
	XSPIN_LOCK(g_iommapSpinLock, flags);

	outb(Block.uData , Block.Port);

	// restore global irq and unlock
	//do { spin_unlock(&g_iommapSpinLock); restore_flags(flags); /*sti();*/}while (0);
	//spin_unlock_irqrestore(&g_iommapSpinLock, flags);
	XSPIN_UNLOCK(g_iommapSpinLock, flags);

	DMSG (KERN_INFO "Block.Port 0x%X, Block.uData 0x%X\n",(unsigned)Block.Port,Block.uData);
	
	return SUCCESSFUL;

}



/********************************************************************************
 * Function	: PciTestReadPortWord												*
 * Access	: Local																*
 * Abstract	: Read IO port register in word.									*
 ********************************************************************************/
static BOOL PciTestReadPortWord(unsigned long pciapiInfo)
{
	int status; unsigned long flags;
	PCITEST_PORT_USHORT_BLOCK Block;

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, pciapiInfo);

	status = copy_from_user( &Block, (void*)pciapiInfo, sizeof(PCITEST_PORT_USHORT_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	// disable global irq and spinlock
	//do { save_flags(flags);	cli(); spin_lock(&g_iommapSpinLock); }while (0);
	//spin_lock_irqsave(&g_iommapSpinLock, flags);
	XSPIN_LOCK(g_iommapSpinLock, flags);

	Block.uData = inw(Block.Port);

	// restore global irq and unlock
	//do { spin_unlock(&g_iommapSpinLock); restore_flags(flags); /*sti();*/}while (0);
	//spin_unlock_irqrestore(&g_iommapSpinLock, flags);
	XSPIN_UNLOCK(g_iommapSpinLock, flags);

	status = copy_to_user((void*)pciapiInfo, &Block, sizeof(PCITEST_PORT_USHORT_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_TO_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	DMSG (KERN_INFO "Block.Port 0x%X, Block.uData 0x%X\n",(unsigned)Block.Port,Block.uData);

    return SUCCESSFUL;
}

/********************************************************************************
 * Function	: PciTestWritePortWord												*
 * Access	: Local																*
 * Abstract	: Write IO port register in word.									*
 ********************************************************************************/
static BOOL PciTestWritePortWord(unsigned long pciapiInfo)
{
	int status; unsigned long flags;
	PCITEST_PORT_USHORT_BLOCK Block;

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, pciapiInfo);

	status = copy_from_user( &Block, (void*)pciapiInfo, sizeof(PCITEST_PORT_USHORT_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	// disable global irq and spinlock
	//do { save_flags(flags);	cli(); spin_lock(&g_iommapSpinLock); }while (0);
	//spin_lock_irqsave(&g_iommapSpinLock, flags);
	XSPIN_LOCK(g_iommapSpinLock, flags);

	outw(Block.uData , Block.Port);

	// restore global irq and unlock
	//do { spin_unlock(&g_iommapSpinLock); restore_flags(flags); /*sti();*/ }while (0);
	//spin_unlock_irqrestore(&g_iommapSpinLock, flags);
	XSPIN_UNLOCK(g_iommapSpinLock, flags);

	DMSG (KERN_INFO "Block.Port 0x%X, Block.uData 0x%X\n",(unsigned)Block.Port,Block.uData);

    return SUCCESSFUL;
}

/********************************************************************************
 * Function	: PciTestReadPortDword												*
 * Access	: Local																*
 * Abstract	: Read IO port register in dword.									*
 ********************************************************************************/
static BOOL PciTestReadPortDword(unsigned long pciapiInfo)
{
	int status; unsigned long flags;
	PCITEST_PORT_ULONG_BLOCK Block;

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, pciapiInfo);

	status = copy_from_user( &Block, (void*)pciapiInfo, sizeof(PCITEST_PORT_ULONG_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	// disable global irq and spinlock
	//do { save_flags(flags);	cli(); spin_lock(&g_iommapSpinLock); }while (0);
	//spin_lock_irqsave(&g_iommapSpinLock, flags);
	XSPIN_LOCK(g_iommapSpinLock, flags);

	Block.uData = inl(Block.Port);

	// restore global irq and unlock
	//do { spin_unlock(&g_iommapSpinLock); restore_flags(flags); /*sti();*/}while (0);
	//spin_unlock_irqrestore(&g_iommapSpinLock, flags);
	XSPIN_UNLOCK(g_iommapSpinLock, flags);

	status = copy_to_user((void*)pciapiInfo, &Block, sizeof(PCITEST_PORT_ULONG_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_TO_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	DMSG (KERN_INFO "Block.Port 0x%X, Block.uData 0x%X\n",(unsigned)Block.Port,Block.uData);

    return SUCCESSFUL;
}

/********************************************************************************
 * Function	: PciTestWritePortDword												*
 * Access	: Local																*
 * Abstract	: Write IO port register in dword.									*
 ********************************************************************************/
static BOOL PciTestWritePortDword(unsigned long pciapiInfo)
{
	int status; unsigned long flags;
	PCITEST_PORT_ULONG_BLOCK Block;

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, pciapiInfo);

	status = copy_from_user( &Block, (void*)pciapiInfo, sizeof(PCITEST_PORT_ULONG_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	// disable global irq and spinlock
	//do { save_flags(flags);	cli(); spin_lock(&g_iommapSpinLock); }while (0);
	//spin_lock_irqsave(&g_iommapSpinLock, flags);
	XSPIN_LOCK(g_iommapSpinLock, flags);

	outl(Block.uData , Block.Port);

	// restore global irq and unlock
	//do { spin_unlock(&g_iommapSpinLock); restore_flags(flags); /*sti();*/}while (0);
	//spin_unlock_irqrestore(&g_iommapSpinLock, flags);
	XSPIN_UNLOCK(g_iommapSpinLock, flags);
	
	DMSG (KERN_INFO "Block.Port 0x%X, Block.uData 0x%X\n",(unsigned)Block.Port,Block.uData);

    return SUCCESSFUL;
}




/********************************************************************************
 * Function	: PciTestReadPortByteEx												*
 * Access	: Local																*
 * Abstract	: Read IO index port register in byte.								*
 ********************************************************************************/
static BOOL PciTestReadPortByteEx(unsigned long pciapiInfo)
{
	int status;	unsigned long flags;
	PCITEST_PORT_UCHAR_EX_BLOCK Block;

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, pciapiInfo);

	status = copy_from_user( &Block, (void*)pciapiInfo, sizeof(PCITEST_PORT_UCHAR_EX_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	// disable global irq and spinlock
	//do { save_flags(flags);	cli(); spin_lock(&g_iommapSpinLock); }while (0);
	//spin_lock_irqsave(&g_iommapSpinLock, flags);
	XSPIN_LOCK(g_iommapSpinLock, flags);
		
	outb(Block.uAddr, Block.AddrPort);
	Block.uData = inb(Block.DataPort);

	// restore global irq and unlock
	//do { spin_unlock(&g_iommapSpinLock); restore_flags(flags); /*sti();*/}while (0);
	//spin_unlock_irqrestore(&g_iommapSpinLock, flags);
	XSPIN_UNLOCK(g_iommapSpinLock, flags);

	status = copy_to_user((void*)pciapiInfo, &Block, sizeof(PCITEST_PORT_UCHAR_EX_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_TO_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	DMSG(KERN_INFO "Block.AddrPort 0x%X, Block.uAddr 0x%X,  Block.DataPort 0x%X, Block.uData 0x%X\n",
				(unsigned)Block.AddrPort, Block.uAddr, (unsigned)Block.DataPort, Block.uData);

    return SUCCESSFUL;
}

/********************************************************************************
 * Function	: PciTestWritePortByteEx											*
 * Access	: Local																*
 * Abstract	: Write IO index port register in byte.								*
 ********************************************************************************/
static BOOL PciTestWritePortByteEx(unsigned long pciapiInfo)
{
	int status;	unsigned long flags;
	PCITEST_PORT_UCHAR_EX_BLOCK Block;

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, pciapiInfo);

	status = copy_from_user( &Block, (void*)pciapiInfo, sizeof(PCITEST_PORT_UCHAR_EX_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	// disable global irq and spinlock
	//do { save_flags(flags);	cli(); spin_lock(&g_iommapSpinLock); }while (0);
	//spin_lock_irqsave(&g_iommapSpinLock, flags);
	XSPIN_LOCK(g_iommapSpinLock, flags);
	
	outb(Block.uAddr, Block.AddrPort);
	outb(Block.uData, Block.DataPort);

	// restore global irq and unlock
	//do { spin_unlock(&g_iommapSpinLock); restore_flags(flags); /*sti();*/}while (0);
	//spin_unlock_irqrestore(&g_iommapSpinLock, flags);
	XSPIN_UNLOCK(g_iommapSpinLock, flags);

	status = copy_to_user((void*)pciapiInfo, &Block, sizeof(PCITEST_PORT_UCHAR_EX_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_TO_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	DMSG(KERN_INFO "Block.AddrPort 0x%X, Block.uAddr 0x%X,  Block.DataPort 0x%X, Block.uData 0x%X\n",
				(unsigned)Block.AddrPort, Block.uAddr, (unsigned)Block.DataPort, Block.uData);

    return SUCCESSFUL;
}

/********************************************************************************
 * Function	: PciTestReadPortWordEx												*
 * Access	: Local																*
 * Abstract	: Read IO index port register in word.								*
 ********************************************************************************/
static BOOL PciTestReadPortWordEx(unsigned long pciapiInfo)
{
	int status;	unsigned long flags;
	PCITEST_PORT_USHORT_EX_BLOCK Block;

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, pciapiInfo);

	status = copy_from_user( &Block, (void*)pciapiInfo, sizeof(PCITEST_PORT_USHORT_EX_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	// disable global irq and spinlock
	//do { save_flags(flags);	cli(); spin_lock(&g_iommapSpinLock); }while (0);
	//spin_lock_irqsave(&g_iommapSpinLock, flags);
	XSPIN_LOCK(g_iommapSpinLock, flags);

	outw(Block.uAddr, Block.AddrPort);
	Block.uData = inw(Block.DataPort);

	// restore global irq and unlock
	//do { spin_unlock(&g_iommapSpinLock); restore_flags(flags); /*sti();*/}while (0);
	//spin_unlock_irqrestore(&g_iommapSpinLock, flags);
	XSPIN_UNLOCK(g_iommapSpinLock, flags);

	status = copy_to_user((void*)pciapiInfo, &Block, sizeof(PCITEST_PORT_USHORT_EX_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_TO_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	DMSG (KERN_INFO "Block.AddrPort 0x%X, Block.uAddr 0x%X,  Block.DataPort 0x%X, Block.uData 0x%X\n",
				(unsigned)Block.AddrPort, Block.uAddr, (unsigned)Block.DataPort, Block.uData);

    return SUCCESSFUL;
}

/********************************************************************************
 * Function	: PciTestWritePortWordEx											*
 * Access	: Local																*
 * Abstract	: Write IO index port register in word.								*
 ********************************************************************************/
static BOOL PciTestWritePortWordEx(unsigned long pciapiInfo)
{
	int status;	unsigned long flags;
	PCITEST_PORT_USHORT_EX_BLOCK Block;

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, pciapiInfo);

	status = copy_from_user( &Block, (void*)pciapiInfo, sizeof(PCITEST_PORT_USHORT_EX_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	// disable global irq and spinlock
	//do { save_flags(flags);	cli(); spin_lock(&g_iommapSpinLock); }while (0);
	//spin_lock_irqsave(&g_iommapSpinLock, flags);
	XSPIN_LOCK(g_iommapSpinLock, flags);

	outw(Block.uAddr, Block.AddrPort);
	outw(Block.uData, Block.DataPort);

	// restore global irq and unlock
	//do { spin_unlock(&g_iommapSpinLock); restore_flags(flags); /*sti();*/}while (0);
	//spin_unlock_irqrestore(&g_iommapSpinLock, flags);
	XSPIN_UNLOCK(g_iommapSpinLock, flags);

	status = copy_to_user((void*)pciapiInfo, &Block, sizeof(PCITEST_PORT_USHORT_EX_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_TO_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	DMSG (KERN_INFO "Block.AddrPort 0x%X, Block.uAddr 0x%X,  Block.DataPort 0x%X, Block.uData 0x%X\n",
				(unsigned)Block.AddrPort, Block.uAddr, (unsigned)Block.DataPort, Block.uData);

    return SUCCESSFUL;
}

/********************************************************************************
 * Function	: PciTestReadPortDwordEx											*
 * Access	: Local																*
 * Abstract	: Read IO index port register in dword.								*
 ********************************************************************************/
static BOOL PciTestReadPortDwordEx(unsigned long pciapiInfo)
{
	int status;	unsigned long flags;
	PCITEST_PORT_ULONG_EX_BLOCK Block;

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, pciapiInfo);

	status = copy_from_user( &Block, (void*)pciapiInfo, sizeof(PCITEST_PORT_ULONG_EX_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	// disable global irq and spinlock
	//do { save_flags(flags);	cli(); spin_lock(&g_iommapSpinLock); }while (0);
	//spin_lock_irqsave(&g_iommapSpinLock, flags);
	XSPIN_LOCK(g_iommapSpinLock, flags);

	outl(Block.uAddr, Block.AddrPort);
	Block.uData = inl(Block.DataPort);

	// restore global irq and unlock
	//do { spin_unlock(&g_iommapSpinLock); restore_flags(flags); /*sti();*/}while (0);
	//spin_unlock_irqrestore(&g_iommapSpinLock, flags);
	XSPIN_UNLOCK(g_iommapSpinLock, flags);

	status = copy_to_user((void*)pciapiInfo, &Block, sizeof(PCITEST_PORT_ULONG_EX_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_TO_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	DMSG (KERN_INFO "Block.AddrPort 0x%X, Block.uAddr 0x%X,  Block.DataPort 0x%X, Block.uData 0x%X\n",
				(unsigned)Block.AddrPort, Block.uAddr, (unsigned)Block.DataPort, Block.uData);

    return SUCCESSFUL;
}

/********************************************************************************
 * Function	: PciTestWritePortDwordEx											*
 * Access	: Local																*
 * Abstract	: Write IO index port register in dword.							*
 ********************************************************************************/
static BOOL PciTestWritePortDwordEx(unsigned long pciapiInfo)
{
	int status;	unsigned long flags;
	PCITEST_PORT_ULONG_EX_BLOCK Block;

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, pciapiInfo);

	status = copy_from_user( &Block, (void*)pciapiInfo, sizeof(PCITEST_PORT_ULONG_EX_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}
	
	// disable global irq and spinlock
	//do { save_flags(flags);	cli(); spin_lock(&g_iommapSpinLock); }while (0);
	//spin_lock_irqsave(&g_iommapSpinLock, flags);
	XSPIN_LOCK(g_iommapSpinLock, flags);

	outl(Block.uAddr, Block.AddrPort);
	outl(Block.uData, Block.DataPort);

	// restore global irq and unlock
	//do { spin_unlock(&g_iommapSpinLock); restore_flags(flags); /*sti();*/}while (0);
	//spin_unlock_irqrestore(&g_iommapSpinLock, flags);
	XSPIN_UNLOCK(g_iommapSpinLock, flags);

	status = copy_to_user((void*)pciapiInfo, &Block, sizeof(PCITEST_PORT_ULONG_EX_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_TO_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	DMSG (KERN_INFO "Block.AddrPort 0x%X, Block.uAddr 0x%X,  Block.DataPort 0x%X, Block.uData 0x%X\n",
				(unsigned)Block.AddrPort, Block.uAddr, (unsigned)Block.DataPort, Block.uData);

    return SUCCESSFUL;
}





/********************************************************************************
 * Function	: PciTestReadRegisterByte											*
 * Access	: Local																*
 * Abstract	: Read IO memory register in byte.									*
 ********************************************************************************/
static BOOL PciTestReadRegisterByte(unsigned long pciapiInfo)
{
	int status; unsigned long flags; void *baseptr = NULL;
	PCITEST_REGISTER_UCHAR_BLOCK Block;

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, pciapiInfo);

	status = copy_from_user( &Block, (void*)pciapiInfo, sizeof(PCITEST_REGISTER_UCHAR_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	// sizeof(UINT8) + 1 is needed
	// ioremap_nocache(0xFC000000, 1 ) will cause kernel segment fault
	baseptr = ioremap_nocache(Block.Register, (sizeof(UINT8)+1));

	if(baseptr == NULL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: INVALID IO_REMAP_NOCACHE POINTER\n",__FUNCTION__, __LINE__);
		return EFAULT;
	}

	// disable global irq and spinlock
	//do { save_flags(flags);	cli(); spin_lock(&g_iommapSpinLock); }while (0);
	//spin_lock_irqsave(&g_iommapSpinLock, flags);
	XSPIN_LOCK(g_iommapSpinLock, flags);

	Block.uData = readb(baseptr);

	// restore global irq and unlock
	//do { spin_unlock(&g_iommapSpinLock); restore_flags(flags); /*sti();*/}while (0);
	//spin_unlock_irqrestore(&g_iommapSpinLock, flags);
	XSPIN_UNLOCK(g_iommapSpinLock, flags);

	iounmap(baseptr);

	status = copy_to_user((void*)pciapiInfo, &Block, sizeof(PCITEST_REGISTER_UCHAR_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_TO_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	DMSG (KERN_INFO "baseptr 0x%lX, Block.Register 0x%X, Block.uData 0x%X\n",
		(unsigned long)baseptr, (unsigned)Block.Register, Block.uData);

	return SUCCESSFUL;
}

/********************************************************************************
 * Function	: PciTestWriteRegisterByte											*
 * Access	: Local																*
 * Abstract	: Write IO memory register in byte.									*
 ********************************************************************************/
static BOOL PciTestWriteRegisterByte(unsigned long pciapiInfo)
{
	int status; unsigned long flags; void *baseptr = NULL;
	PCITEST_REGISTER_UCHAR_BLOCK Block;

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, pciapiInfo);

	status = copy_from_user( &Block, (void*)pciapiInfo, sizeof(PCITEST_REGISTER_UCHAR_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	// sizeof(UINT8) + 1 is needed
	// ioremap_nocache(0xFC000000, 1 ) will cause kernel segment fault
	baseptr = ioremap_nocache(Block.Register, (sizeof(UINT8)+1));

	if(baseptr == NULL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: INVALID IO_REMAP_NOCACHE POINTER\n",__FUNCTION__, __LINE__);
		return EFAULT;
	}

	// disable global irq and spinlock
	//do { save_flags(flags);	cli(); spin_lock(&g_iommapSpinLock); }while (0);
	//spin_lock_irqsave(&g_iommapSpinLock, flags);
	XSPIN_LOCK(g_iommapSpinLock, flags);

	writeb(Block.uData, baseptr);

	// restore global irq and unlock
	//do { spin_unlock(&g_iommapSpinLock); restore_flags(flags); /*sti();*/}while (0);
	//spin_unlock_irqrestore(&g_iommapSpinLock, flags);
	XSPIN_UNLOCK(g_iommapSpinLock, flags);

	iounmap(baseptr);

	DMSG (KERN_INFO "baseptr 0x%lX, Block.Register 0x%X, Block.uData 0x%X\n",
		(unsigned long)baseptr, (unsigned)Block.Register,Block.uData);

    return SUCCESSFUL;
}


/********************************************************************************
 * Function	: PciTestReadRegisterWord											*
 * Access	: Local																*
 * Abstract	: Read IO memory register in word.									*
 ********************************************************************************/
static BOOL PciTestReadRegisterWord(unsigned long pciapiInfo)
{
	int status; unsigned long flags; void *baseptr = NULL;
	PCITEST_REGISTER_USHORT_BLOCK Block;

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, pciapiInfo);

	status = copy_from_user( &Block, (void*)pciapiInfo, sizeof(PCITEST_REGISTER_USHORT_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	baseptr = ioremap_nocache(Block.Register, sizeof(UINT16));

	if(baseptr == NULL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: INVALID IO_REMAP_NOCACHE POINTER\n",__FUNCTION__, __LINE__);
		return EFAULT;
	}

	// disable global irq and spinlock
	//do { save_flags(flags);	cli(); spin_lock(&g_iommapSpinLock); }while (0);
	//spin_lock_irqsave(&g_iommapSpinLock, flags);
	XSPIN_LOCK(g_iommapSpinLock, flags);

	Block.uData = readw(baseptr);

	// restore global irq and unlock
	//do { spin_unlock(&g_iommapSpinLock); restore_flags(flags); /*sti();*/}while (0);
	//spin_unlock_irqrestore(&g_iommapSpinLock, flags);
	XSPIN_UNLOCK(g_iommapSpinLock, flags);

	iounmap(baseptr);

	status = copy_to_user((void*)pciapiInfo, &Block, sizeof(PCITEST_REGISTER_USHORT_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_TO_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	DMSG (KERN_INFO "baseptr 0x%lX, Block.Register 0x%X, Block.uData 0x%X\n",
		(unsigned long)baseptr, (unsigned)Block.Register,Block.uData);

    return SUCCESSFUL;
}

/********************************************************************************
 * Function	: PciTestWriteRegisterWord											*
 * Access	: Local																*
 * Abstract	: Write IO memory register in word.									*
 ********************************************************************************/
static BOOL PciTestWriteRegisterWord(unsigned long pciapiInfo)
{
	int status; unsigned long flags; void *baseptr = NULL;
	PCITEST_REGISTER_USHORT_BLOCK Block;

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, pciapiInfo);

	status = copy_from_user( &Block, (void*)pciapiInfo, sizeof(PCITEST_REGISTER_USHORT_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	baseptr = ioremap_nocache(Block.Register, sizeof(UINT16));

	if(baseptr == NULL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: INVALID IO_REMAP_NOCACHE POINTER\n",__FUNCTION__, __LINE__);
		return EFAULT;
	}

	// disable global irq and spinlock
	//do { save_flags(flags);	cli(); spin_lock(&g_iommapSpinLock); }while (0);
	//spin_lock_irqsave(&g_iommapSpinLock, flags);
	XSPIN_LOCK(g_iommapSpinLock, flags);

	writew(Block.uData, baseptr);

	// restore global irq and unlock
	//do { spin_unlock(&g_iommapSpinLock); restore_flags(flags); /*sti();*/}while (0);
	//spin_unlock_irqrestore(&g_iommapSpinLock, flags);
	XSPIN_UNLOCK(g_iommapSpinLock, flags);

	iounmap(baseptr);

	DMSG (KERN_INFO "baseptr 0x%lX, Block.Register 0x%X, Block.uData 0x%X\n",
		(unsigned long)baseptr, (unsigned)Block.Register,Block.uData);

    return SUCCESSFUL;
}

/********************************************************************************
 * Function	: PciTestReadRegisterDword											*
 * Access	: Local																*
 * Abstract	: Read IO memory register in dword.									*
 ********************************************************************************/
static BOOL PciTestReadRegisterDword(unsigned long pciapiInfo)
{
	int status; unsigned long flags; void *baseptr = NULL;
	PCITEST_REGISTER_ULONG_BLOCK Block;

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, pciapiInfo);

	status = copy_from_user( &Block, (void*)pciapiInfo, sizeof(PCITEST_REGISTER_ULONG_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}
	
	baseptr = ioremap_nocache(Block.Register, sizeof(UINT32));

	if(baseptr == NULL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: INVALID IO_REMAP_NOCACHE POINTER\n",__FUNCTION__, __LINE__);
		return EFAULT;
	}

	// disable global irq and spinlock
	//do { save_flags(flags);	cli(); spin_lock(&g_iommapSpinLock); }while (0);
	//spin_lock_irqsave(&g_iommapSpinLock, flags);
	XSPIN_LOCK(g_iommapSpinLock, flags);

	Block.uData = readl(baseptr);

	// restore global irq and unlock
	//do { spin_unlock(&g_iommapSpinLock); restore_flags(flags); /*sti();*/}while (0);
	//spin_unlock_irqrestore(&g_iommapSpinLock, flags);
	XSPIN_UNLOCK(g_iommapSpinLock, flags);

	iounmap(baseptr);

	status = copy_to_user((void*)pciapiInfo, &Block, sizeof(PCITEST_REGISTER_ULONG_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_TO_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	DMSG (KERN_INFO "baseptr 0x%lX, Block.Register 0x%X, Block.uData 0x%X\n",
		(unsigned long)baseptr, (unsigned)Block.Register,Block.uData);

    return SUCCESSFUL;
}

/********************************************************************************
 * Function	: PciTestWriteRegisterDword											*
 * Access	: Local																*
 * Abstract	: Write IO memory register in dword.								*
 ********************************************************************************/
static BOOL PciTestWriteRegisterDword(unsigned long pciapiInfo)
{
	int status; unsigned long flags; void *baseptr = NULL;
	PCITEST_REGISTER_ULONG_BLOCK Block;

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, pciapiInfo);

	status = copy_from_user( &Block, (void*)pciapiInfo, sizeof(PCITEST_REGISTER_ULONG_BLOCK));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	baseptr = ioremap_nocache(Block.Register, sizeof(UINT32));

	if(baseptr == NULL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: INVALID IO_REMAP_NOCACHE POINTER\n",__FUNCTION__, __LINE__);
		return EFAULT;
	}

	// disable global irq and spinlock
	//do { save_flags(flags);	cli(); spin_lock(&g_iommapSpinLock); }while (0);
	//spin_lock_irqsave(&g_iommapSpinLock, flags);
	XSPIN_LOCK(g_iommapSpinLock, flags);

	writel(Block.uData, baseptr);

	// restore global irq and unlock
	//do { spin_unlock(&g_iommapSpinLock); restore_flags(flags); /*sti();*/}while (0);
	//spin_unlock_irqrestore(&g_iommapSpinLock, flags);
	XSPIN_UNLOCK(g_iommapSpinLock, flags);

	iounmap(baseptr);

	DMSG (KERN_INFO "baseptr 0x%lX, Block.Register 0x%X, Block.uData 0x%X\n",
		(unsigned long)baseptr, (unsigned)Block.Register,Block.uData);

    return SUCCESSFUL;
}



