////////////////////////////////////////////////////////////////////////////
//  INTEL CONFIDENTIAL.  Copyright (c) 2003-2007 Intel Corp.              //
//  All rights reserved.                                                  // 
////////////////////////////////////////////////////////////////////////////
#ifndef __TDEF_H__
#define __TDEF_H__

/*****************************************************************************\
	Info - tdef.h revamped 4/2005  
		- Added typedef requirments for _NALTYPES_H_ - LAD group sdk for nic testing
		- Corrected some definitions to be os compliant - may break some projects
		- Seperated sections for os specific defines for easier maintenance
		- 64 bit defines not included yet
\*****************************************************************************/



/*****************************************************************************\
	LINUX DEFINITIONS - mutually exclusive of _WIN32, EFI & DOS sections
\*****************************************************************************/
#ifdef linux
#ifndef _WIN32
#ifndef MSDOS
#ifndef EFI

#ifndef _DRIVER_BUILD
#include <sys/types.h>
#endif

#ifdef _DEBUG
#	define STATIC
#else
#	define STATIC static
#endif



#ifndef _NALTYPES_H_	// naltypes is part of NIC low level library
#	ifndef ASSERT
#		define ASSERT(x) assert(x)
#	endif
#endif
#define _ASSERT(x) assert(x)
#define _ASSERTE(x) assert(x)

#define CONST	const
#define TRUE	1
#define FALSE	0


#ifdef _NALTYPES_H_	// naltypes is part of NIC low level library
// 11/23/2005 (kps) this changed to correlate with the change I made to the SDK in naltypes.h in order to compile 
// successfully.  Mark Montecalvo will maybe make the change permanent.  He is considering it.
//	typedef unsigned char	CHAR,	*PCHAR;			
	typedef char	CHAR,	*PCHAR;			
#else
	typedef char		CHAR,	*PCHAR;			
#endif


typedef long int			INTN,	*PINTN;
typedef unsigned long int	UINTN,	*PUINTN;

typedef signed char			INT8,	*PINT8;
typedef unsigned char		UINT8,	*PUINT8;
typedef short				INT16,	*PINT16;
typedef unsigned short		UINT16,	*PUINT16;
typedef int					INT32,	*PINT32;
typedef unsigned int		UINT32,	*PUINT32;
typedef long long			INT64,	*PINT64;
typedef unsigned long long	UINT64,	*PUINT64;

typedef unsigned int		UINT,	*PUINT;
typedef unsigned long		ULONG,	*PULONG;

typedef unsigned char		BYTE,	*PBYTE;
typedef unsigned char		UCHAR,	*PUCHAR;
typedef unsigned short		WORD,	*PWORD;
typedef unsigned int		DWORD,	*PDWORD;

typedef int					BOOL,	*PBOOL;
typedef void				VOID,	*LPVOID;
typedef const void*			LPCVOID;
typedef pid_t				PID,	*PPID;


typedef signed char			CHAR8,	*PCHAR8;
typedef unsigned short		CHAR16,	*PCHAR16;
typedef char*				LPSTR;
typedef const char*			LPCTSTR;
typedef const char*			PCCH;
// tchar stuff defined so that will work for dual compiles with windows 
//   - unicode is not directly supported in linux like this.
typedef char				TCHAR,	TSTR;
typedef char*				LPTSTR;
typedef const char*			LPCSTR;

typedef unsigned long long  TESTHANDLE;


#endif //!EFI
#endif //!MSDOS
#endif //!_WIN32
#endif //linux
///////////////////////////////////////////////////////////////////////////////




/*****************************************************************************\
	_WIN32 DEFINITIONS - mutually exclusive of LINUX, EFI & MSDOS sections
\*****************************************************************************/
#ifdef _WIN32
#ifndef linux
#ifndef MSDOS
#ifndef EFI

typedef int		BOOL,	*PBOOL;
#define CONST	const
#define TRUE	1
#define FALSE	0

#ifdef _DEBUG
#	define STATIC
#else
#	define STATIC static
#endif


#ifndef _DRIVER_BUILD
#include <sys/types.h>
#endif 
	
#ifdef _NALTYPES_H_
#else
#endif

typedef unsigned long int	UINTN,	*PUINTN;
typedef long int			INTN,	*PINTN;

typedef unsigned char		UINT8,	*PUINT8;
typedef signed char			INT8,	*PINT8;

typedef short				INT16,	*PINT16;
typedef unsigned short		UINT16,	*PUINT16;


typedef char				CHAR8,	*PCHAR8;

// These types removed because they are in included in the latest platform sdk.  
//  DO not redefine these types in your project, use the sdk definitions.
//typedef const char*		LPCTSTR;
//typedef unsigned char		BYTE,	*PBYTE;
//typedef char				CHAR,	*PCHAR;
//typedef unsigned short	WORD,	*PWORD;
//typedef unsigned long		DWORD,	*PDWORD;
//typedef unsigned short	UINT16,	*PUINT16;
//typedef unsigned int		UINT32,	*PUINT32;

// socklen_t is used by linux but not defined in Windows - this allows code to dual compile
typedef int socklen_t;


typedef unsigned __int64  TESTHANDLE;


#endif //!EFI
#endif// MSDOS
#endif// linux
#endif// _WIN32
///////////////////////////////////////////////////////////////////////////////


/*****************************************************************************\
	DOS DEFINITIONS - mutually exclusive of _WIN32, EFI & LINUX sections
\*****************************************************************************/
#ifdef MSDOS
#ifndef linux
#ifndef _WIN32
#ifndef EFI

#define CONST	const
#define TRUE	1
#define FALSE	0

typedef signed char		INT8,	*PINT8;
typedef unsigned long   UINT32,	*PUINT32;
typedef long            INT32,	*PINT32;
typedef unsigned int	UINT;
typedef unsigned int	UINTN,	*PUINTN;
typedef int				INTN,	*PINTN;
typedef void			VOID,	*LPVOID;

typedef int					BOOL,	*PBOOL;
typedef unsigned char		BYTE,	*PBYTE;							
typedef unsigned char		UINT8,	*PUINT8;
typedef short				INT16,	*PINT16;
typedef unsigned short		UINT16,	*PUINT16;
typedef unsigned short		WORD,	*PWORD;
typedef unsigned long		DWORD,	*PDWORD;


typedef unsigned char		UCHAR,	*PUCHAR;
typedef signed char			CHAR8,	*PCHAR8;
typedef short 				CHAR16,	*PCHAR16;
typedef const signed char*	LPCSTR;
typedef signed char*		LPSTR;
typedef signed char			TCHAR,	TSTR;
typedef signed char*		LPTSTR;


#ifdef _NALTYPES_H_
	typedef unsigned char			CHAR,	*PCHAR;
	typedef const char*				LPCTSTR;
#else
	typedef signed char				CHAR,	*PCHAR;
	typedef const char*		LPCTSTR;
#endif


#endif //!EFI
#endif //!_WIN32
#endif //!linux
#endif //MSDOS
///////////////////////////////////////////////////////////////////////////////

/*****************************************************************************\
	EFI DEFINITIONS - mutually exclusive of DOS, _WIN32 & LINUX sections
\*****************************************************************************/
#ifdef EFI
#ifndef MSDOS
#ifndef linux
#ifndef _WIN32

// efi typedefs go here

#endif //!_WIN32
#endif //!linux
#endif //!MSDOS
#endif // EFI
///////////////////////////////////////////////////////////////////////////////


/*****************************************************************************\
	General struct and application definitions 
		- MAINTAIN BASIC TYPES IN OS SPECIFIC SECTIONS ABOVE
\*****************************************************************************/

// convenience define
enum binaryNumbers{
	b00,
	b01,
	b10,
	b11
};
///////////////////////////////////////////////////////////////////////////////


	
typedef enum eTSSTATUS
{
	TSE_SUCCESS,
	TSE_UNSUPPORTED,
	TSE_UNSUCCESSFUL,
	TSE_UNINITIALIZED,
	TSE_NOT_FOUND,
	TSE_INVALID_PARAMETER,
	TSE_BUFFER_TOO_SMALL,
	TSE_OUT_OF_RANGE,
	TSE_OUT_OF_RESOURCES,
	TSE_NOT_AUTOCONFIG,
	TSE_NOT_AVAILABLE,
	TSE_DATA_INCOMPLETE,
	TSE_ERROR_RETURNED,
	TSE_EXTRA_DATA,
	TSE_TIMEOUT,
	TSE_BAD_ACK,
	TSE_CANCEL,
	TSE_RETRY,
	TSE_FAILED_MFC_INITIALIZATION,
	TSE_FAILED_WT_INITIALIZATION,
	TSE_FAILED_TS_INITIALIZATION,
	TSE_LICENSE_FAILED,

} TSSTATUS;
///////////////////////////////////////////////////////////////////////////////


#endif  //__TDEF_H__

