/*
Copyright (c) 2017, Intel Corp.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of Intel Corporation nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef WRAPPER_H
#define WRAPPER_H

#include <linux/version.h>

/********************************************************************************
 * General for debug or release build											*
 ********************************************************************************/
#ifdef _DEBUG
#define	DMSG(fmt, arg...) printk(fmt, ##arg)
#define BUILD "debug"
#define XINIT_VMA(g_vma, g_vma_offset) do {g_vma = vma; g_vma_offset = (unsigned long)Block[0].uAddress; }while(0)
#else
#define	DMSG(fmt, arg...) do { } while(0)
#define BUILD "release"
#define XINIT_VMA(g_vma, g_vma_offset) do {g_vma = NULL; g_vma_offset = 0; }while(0)
#endif

/********************************************************************************
 * if x86_64 or m686, for machine string compat									*
 ********************************************************************************/
//#ifdef CONFIG_X86_64
#ifdef x86_64
#define MACHINE "x86_64"
//#elif CONFIG_M686
#elif defined(i386)
#define MACHINE "i386"
#elif defined(ia64)
#define MACHINE "ia64"
#else
#define MACHINE "UNKNOW MACHINE"
#endif

/********************************************************************************
 * if x86_64 or m686, for ioctl compat											*
 * AS3.0, kernel 2.4 - no typdef for ioctl_trans_handler_t						*
 ********************************************************************************/
#ifdef CONFIG_COMPAT
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
typedef int (*ioctl_trans_handler_t)(unsigned int, unsigned int, unsigned long, struct file *);
#endif
#endif

/********************************************************************************
 * Linux 2.4 and Linux 2.6 compat												*
 ********************************************************************************/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#define	MODULE_PARM_INT(name, type, perm) MODULE_PARM(name, type)
#define MODULE_VERSION(_version) \
const char __module_version[] __attribute__((section(".modinfo"))) = "version=" _version
#define IOMMAP_MODULE_INCREASE MOD_INC_USE_COUNT
#define IOMMAP_MODULE_DECREASE MOD_DEC_USE_COUNT
#define PCIBIOS_PRESENT() pcibios_present()
#define XSPIN_LOCK(lock, flags) do { save_flags(flags); cli(); spin_lock(&lock); }while (0)
#define XSPIN_UNLOCK(lock, flags) do { spin_unlock(&lock); restore_flags(flags); /*sti();*/ }while (0)
#define VMA_NOPAGE_WRAPPER(vma, address, write) \
struct page *driver_vma_nopage(struct vm_area_struct *vma, unsigned long address, int write)
/* Linux 2.6 */
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#define	MODULE_PARM_INT(name, type, perm) module_param(name, int, perm)
#define IOMMAP_MODULE_INCREASE try_module_get(THIS_MODULE)
#define IOMMAP_MODULE_DECREASE module_put(THIS_MODULE)
#define PCIBIOS_PRESENT() 1
#define XSPIN_LOCK(lock, flags) \
do { \
	local_irq_save(flags); \
	local_irq_disable(); \
	spin_lock(&lock); \
} while (0)	

#define XSPIN_UNLOCK(lock, flags) \
do { \
	spin_unlock(&lock); \
	local_irq_restore(flags); \
} while (0)	

//#define XSPIN_LOCK(lock, flags) spin_lock_irqsave(&lock, flags)
//#define XSPIN_UNLOCK(lock, flags) spin_unlock_irqrestore(&lock, flags)
#define VMA_NOPAGE_WRAPPER(vma, address, write) \
struct page *driver_vma_nopage(struct vm_area_struct *vma, unsigned long address, int *write)
#endif

/********************************************************************************
 * Linux 2.4 and Linux 2.6 PCI legacy		  									*
 ********************************************************************************/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
#define PCI_FIND_SLOT(bus, devfn) pci_find_slot(bus, devfn)
#elif LINUX_VERSION_CODE > KERNEL_VERSION(3,0,0) & LINUX_VERSION_CODE < KERNEL_VERSION(4,11,99)
#define PCI_FIND_SLOT(bus, devfn) pci_get_bus_and_slot(bus, devfn)
#elif LINUX_VERSION_CODE > KERNEL_VERSION(4,15,0)
#define PCI_FIND_SLOT(bus, devfn) pci_get_domain_bus_and_slot(0, bus, devfn)
#else
#ifdef CONFIG_PCI_LEGACY
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,32)
#define PCI_FIND_SLOT(bus, devfn) pci_find_slot(bus, devfn)
#else
#define PCI_FIND_SLOT(bus, devfn) pci_get_bus_and_slot(bus, devfn)
#endif //#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,32)
#else
#define PCI_FIND_SLOT(bus, devfn) pci_get_bus_and_slot(bus, devfn)
#endif //#ifdef CONFIG_PCI_LEGACY
#endif //#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)


/********************************************************************************
 * Redhat compat																*
 * Redhat 8.0 - pte_offset, none												*
 * Redhat 9.0, Redhat AS3.2 - pte_offset_map, pte_unmap							*
 ********************************************************************************/
#ifdef _REDHAT
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,20)
#define PUD_DEFINE
#define PGD_OR_PUD pgd
#define PUD_OFFSET_WRAPPER(pgd, address) do { } while(0)
#define PTE_OFFSET_WRAPPER(dir, address) pte_offset(dir, address)
#define PTE_UNMAP_WRAPPER(pte) do { } while(0)

#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) & LINUX_VERSION_CODE < KERNEL_VERSION(2,6,12)
#define PUD_DEFINE
#define PGD_OR_PUD pgd
#define PUD_OFFSET_WRAPPER(pgd, address) do { } while(0)
#define PTE_OFFSET_WRAPPER(dir, address) pte_offset_map(dir, address)
#define PTE_UNMAP_WRAPPER(pte) pte_unmap(pte)

#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12) & LINUX_VERSION_CODE < KERNEL_VERSION(2,7,0)
#define PUD_DEFINE pud_t *pud = NULL;
#define PGD_OR_PUD pud
#define PUD_OFFSET_WRAPPER(pgd, address) pud = pud_offset(pgd, address)
#define PTE_OFFSET_WRAPPER(dir, address) pte_offset_map(dir, address)
#define PTE_UNMAP_WRAPPER(pte) pte_unmap(pte)

#elif LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0) & LINUX_VERSION_CODE < KERNEL_VERSION(3,99,99)
#define PUD_DEFINE pud_t *pud = NULL;
#define PGD_OR_PUD pud
#define PUD_OFFSET_WRAPPER(pgd, address) pud = pud_offset(pgd, address)
#define PTE_OFFSET_WRAPPER(dir, address) pte_offset_map(dir, address)
#define PTE_UNMAP_WRAPPER(pte) pte_unmap(pte)

#elif LINUX_VERSION_CODE >= KERNEL_VERSION(4,0,0) & LINUX_VERSION_CODE < KERNEL_VERSION(5,99,99)
#define PUD_DEFINE pud_t *pud = NULL;
#define PGD_OR_PUD pud
#define PUD_OFFSET_WRAPPER(pgd, address) pud = pud_offset(pgd, address)
#define PTE_OFFSET_WRAPPER(dir, address) pte_offset_map(dir, address)
#define PTE_UNMAP_WRAPPER(pte) pte_unmap(pte)
#else
#define PUD_DEFINE
#define PGD_OR_PUD pgd
#define PUD_OFFSET_WRAPPER(pgd, address) do { } while(0)
#define PTE_OFFSET_WRAPPER(dir, address) pte_offset_map(dir, address)
#define PTE_UNMAP_WRAPPER(pte) do { } while(0)
#endif
#endif

/********************************************************************************
 * SuSE compat																	*
 * SuSE 8.2, SuSE 9.0 - pte_offset_atomic, pte_kunmap							*
 * SuSE 10.0		  - pud_offset												*
 ********************************************************************************/
#ifdef _SUSE
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,22)
#define PUD_DEFINE
#define PGD_OR_PUD pgd
#define PUD_OFFSET_WRAPPER(pgd, address) do { } while(0)
#define PTE_OFFSET_WRAPPER(dir, address) pte_offset_atomic(dir, address)
#define PTE_UNMAP_WRAPPER(pte) pte_kunmap(pte)

#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) & LINUX_VERSION_CODE < KERNEL_VERSION(2,6,12)
#define PUD_DEFINE
#define PGD_OR_PUD pgd
#define PUD_OFFSET_WRAPPER(pgd, address) do { } while(0)
#define PTE_OFFSET_WRAPPER(dir, address) pte_offset_map(dir, address)
#define PTE_UNMAP_WRAPPER(pte) pte_unmap(pte)

#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12) & LINUX_VERSION_CODE < KERNEL_VERSION(2,7,0)
#define PUD_DEFINE pud_t *pud = NULL;
#define PGD_OR_PUD pud
#define PUD_OFFSET_WRAPPER(pgd, address) pud = pud_offset(pgd, address)
#define PTE_OFFSET_WRAPPER(dir, address) pte_offset_map(dir, address)
#define PTE_UNMAP_WRAPPER(pte) pte_unmap(pte)

#elif LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0) & LINUX_VERSION_CODE < KERNEL_VERSION(3,99,99)
#define PUD_DEFINE pud_t *pud = NULL;
#define PGD_OR_PUD pud
#define PUD_OFFSET_WRAPPER(pgd, address) pud = pud_offset(pgd, address)
#define PTE_OFFSET_WRAPPER(dir, address) pte_offset_map(dir, address)
#define PTE_UNMAP_WRAPPER(pte) pte_unmap(pte)

#elif LINUX_VERSION_CODE >= KERNEL_VERSION(4,0,0) & LINUX_VERSION_CODE < KERNEL_VERSION(5,99,99)
#define PUD_DEFINE pud_t *pud = NULL;
#define PGD_OR_PUD pud
#define PUD_OFFSET_WRAPPER(pgd, address) pud = pud_offset(pgd, address)
#define PTE_OFFSET_WRAPPER(dir, address) pte_offset_map(dir, address)
#define PTE_UNMAP_WRAPPER(pte) pte_unmap(pte)
#else
#define PUD_DEFINE
#define PGD_OR_PUD pgd
#define PUD_OFFSET_WRAPPER(pgd, address) do { } while(0)
#define PTE_OFFSET_WRAPPER(dir, address) do { } while(0)
#define PTE_UNMAP_WRAPPER(pte) do { } while(0)
#endif
#endif

/********************************************************************************
 * MVL CGE3.1 compat															*
 * MontaVista CGE 3.1 - pte_offset, none										*
 ********************************************************************************/
#ifdef _MVL
#define PUD_DEFINE
#define PGD_OR_PUD pgd
#define PUD_OFFSET_WRAPPER(pgd, address) do { } while(0)
#define PTE_OFFSET_WRAPPER(dir, address) pte_offset(dir, address)
#define PTE_UNMAP_WRAPPER(pte) do { } while(0)
#endif

/********************************************************************************
 * Gentoo compat																*
 * Gentoo 1.4 - pte_offset, none												*
 ********************************************************************************/
#ifdef _GENTOO
#define PUD_DEFINE
#define PGD_OR_PUD pgd
#define PUD_OFFSET_WRAPPER(pgd, address) do { } while(0)
#define PTE_OFFSET_WRAPPER(dir, address) pte_offset(dir, address)
#define PTE_UNMAP_WRAPPER(pte) do { } while(0)
#endif

/********************************************************************************
 * LFS compat																	*
 * LFS - pte_offset, none														*
 ********************************************************************************/
#ifdef _LFS
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,22)
#define PUD_DEFINE
#define PGD_OR_PUD pgd
#define PUD_OFFSET_WRAPPER(pgd, address) do { } while(0)
#define PTE_OFFSET_WRAPPER(dir, address) pte_offset_atomic(dir, address)
#define PTE_UNMAP_WRAPPER(pte) pte_kunmap(pte)
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,7,0)
#define PUD_DEFINE
#define PGD_OR_PUD pgd
#define PUD_OFFSET_WRAPPER(pgd, address) do { } while(0)
#define PTE_OFFSET_WRAPPER(dir, address) pte_offset_map(dir, address)
#define PTE_UNMAP_WRAPPER(pte) pte_unmap(pte)
#endif
#endif

/********************************************************************************
 * Ubuntu compat																*
 * Redhat 8.0 - pte_offset, none												*
 * Redhat 9.0, Redhat AS3.2 - pte_offset_map, pte_unmap							*
 ********************************************************************************/
#ifdef _UBUNTU
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,20)
#define PUD_DEFINE
#define PGD_OR_PUD pgd
#define PUD_OFFSET_WRAPPER(pgd, address) do { } while(0)
#define PTE_OFFSET_WRAPPER(dir, address) pte_offset(dir, address)
#define PTE_UNMAP_WRAPPER(pte) do { } while(0)

#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) & LINUX_VERSION_CODE < KERNEL_VERSION(2,6,12)
#define PUD_DEFINE
#define PGD_OR_PUD pgd
#define PUD_OFFSET_WRAPPER(pgd, address) do { } while(0)
#define PTE_OFFSET_WRAPPER(dir, address) pte_offset_map(dir, address)
#define PTE_UNMAP_WRAPPER(pte) pte_unmap(pte)

#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12) & LINUX_VERSION_CODE < KERNEL_VERSION(2,7,0)
#define PUD_DEFINE pud_t *pud = NULL;
#define PGD_OR_PUD pud
#define PUD_OFFSET_WRAPPER(pgd, address) pud = pud_offset(pgd, address)
#define PTE_OFFSET_WRAPPER(dir, address) pte_offset_map(dir, address)
#define PTE_UNMAP_WRAPPER(pte) pte_unmap(pte)

#elif LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0) & LINUX_VERSION_CODE < KERNEL_VERSION(3,99,99)
#define PUD_DEFINE pud_t *pud = NULL;
#define PGD_OR_PUD pud
#define PUD_OFFSET_WRAPPER(pgd, address) pud = pud_offset(pgd, address)
#define PTE_OFFSET_WRAPPER(dir, address) pte_offset_map(dir, address)
#define PTE_UNMAP_WRAPPER(pte) pte_unmap(pte)

#elif LINUX_VERSION_CODE >= KERNEL_VERSION(4,0,0) & LINUX_VERSION_CODE < KERNEL_VERSION(5,99,99)
#define PUD_DEFINE pud_t *pud = NULL;
#define PGD_OR_PUD pud
#define PUD_OFFSET_WRAPPER(pgd, address) pud = pud_offset(pgd, address)
#define PTE_OFFSET_WRAPPER(dir, address) pte_offset_map(dir, address)
#define PTE_UNMAP_WRAPPER(pte) pte_unmap(pte)
#else
#define PUD_DEFINE
#define PGD_OR_PUD pgd
#define PUD_OFFSET_WRAPPER(pgd, address) do { } while(0)
#define PTE_OFFSET_WRAPPER(dir, address) pte_offset_map(dir, address)
#define PTE_UNMAP_WRAPPER(pte) do { } while(0)
#endif
#endif

/********************************************************************************
 * No OS compat																	*
 * Check for the unknown OS														*
 ********************************************************************************/
#ifndef _UNDEF
#define CHECK_UNKNOW_OS_ERROR_1() do { }while(0)
#define CHECK_UNKNOW_OS_ERROR_2() do { }while(0)
#endif
#ifdef _UNDEF
#define CHECK_UNKNOW_OS_ERROR_1() \
do { \
	g_vma = NULL; \
	g_vma_offset = 0; \
	printk (KERN_ERR "ERROR UNSUPPORTED OPERATING SYSTEM VERSION\n"); \
	return ENOSYS; \
}while(0)

#define CHECK_UNKNOW_OS_ERROR_2() \
do { \
	printk (KERN_ERR "ERROR UNSUPPORTED OPERATING SYSTEM VERSION\n"); \
	return page; \
}while(0)
#endif


#endif /* #endif WRAPPER_H*/

